"""
Created on Jul 10, 2011

Implements the structures used by the ANAPSID Operators.

@author: Maribel Acosta Deibe
"""

# from time import time


class Record(object):
    """
    Represents a structure that is inserted into the hash table.
    It is composed by a tuple, probeTS (timestamp when the tuple was probed)
    and insertTS (timestamp when the tuple was inserted in the table).
    """

    def __init__(self, tuple, probeTS, insertTS=None, flushTS=None):
        self.tuple = tuple
        self.probeTS = probeTS
        self.insertTS = insertTS
        self.flushTS = flushTS


class RJTTail(object):
    """
    Represents the tail of a RJT.
    It is composed by a list of records and rjtprobeTS
    (timestamp when the last tuple in the RJT was probed).
    """

    def __init__(self, record, rjtProbeTS):
        self.records = [record]
        self.rjtProbeTS = rjtProbeTS
        self.flushTS = float("inf")

    def updateRecords(self, record):
        self.records.append(record)

    def setRJTProbeTS(self, rjtProbeTS):
        self.rjtProbeTS = rjtProbeTS


class FileDescriptor(object):
    """
    Represents the description of a file, that contains a RJT in sec mem.
    It is composed by the name of the file, the associated resource,
    the current size (number of tuples), and the timestamp of the last
    RJTTail that have been flushed.
    """

    def __init__(self, file, size, lastFlushTS):
        self.file = file
        self.size = size
        self.lastFlushTS = lastFlushTS
        # self.table = table

    def getSize(self):
        return self.size

#    def setSize(self, size):
#        self.size = size
#
#    def setLastFlushTS(self, lastFlushTS):
#        self.lastFlushTS = lastFlushTS



"""
Created on Jul 10, 2011

Implements the Xlimit operator.
The intermediate results are represented in a queue.

@author: Maribel Acosta Deibe
"""

from multiprocessing import Queue


class Xlimit(object):

    name = "LIMIT"

    def __init__(self, vars, limit):
        self.input = Queue()
        self.qresults = Queue()
        self.vars = vars
        self.limit = int(limit)

    def execute(self, left, dummy, out, processqueue=Queue()):
        # print  "Executes the Xlimit.", self.limit
        self.left = left
        self.qresults = out
        tuple = self.left.get(True)
        count = 0

        # LIMIT.
        while ((count < self.limit) and (tuple != "EOF")):
            self.qresults.put(tuple)
            count = count + 1
            # print "count", count
            tuple = self.left.get(True)

        # Put EOF in queue and exit.
        self.qresults.put("EOF")
        return



import abc
from multiprocessing import Queue


class _Union(object):
    __metaclass__ = abc.ABCMeta

    name = "UNION"

    @abc.abstractmethod
    def execute(self, left, right, out, processqueue=Queue()):
        return

    @abc.abstractmethod
    def instantiate(self, d):
        return

    def getCardinality(self, l, r):
        c1 = l.getCardinality()
        c2 = r.getCardinality()
        return c1 + c2

    def getJoinCardinality(self, l, r, vars):
        return l.getJoinCardinality(vars) + r.getJoinCardinality(vars)



from multiprocessing import Queue
from DeTrusty.Sparql.Parser.services import Bind, Expression, Argument
import math
import datetime
import operator


arithmetic_operators = {
    '*': operator.mul,
    '/': operator.truediv,
    '+': operator.add,
    '-': operator.sub,
}

unary_operators = {
    '!': operator.not_,
    'bound': lambda a: len(a) > 0,
    'str': str,
    'floor': math.floor,
    'year': datetime.datetime.year
}

logical_connectives = {
    '||': operator.or_,
    '&&': operator.and_
}

test_operators = {
    '=': operator.eq,
    '!=': operator.ne,
    '<': operator.lt,
    '>': operator.gt,
    '<=': operator.le,
    '>=': operator.ge
}

data_types = {
    'integer': (int, 'numerical'),
    'decimal': (float, 'numerical'),
    'float': (float, 'numerical'),
    'double': (float, 'numerical'),
    'string': (str, str),
    'boolean': (bool, bool),
    'dateTime': (datetime, datetime),
    'nonPositiveInteger': (int, 'numerical'),
    'negativeInteger': (int, 'numerical'),
    'long': (int, 'numerical'),
    'int': (int, 'numerical'),
    'short': (int, 'numerical'),
    'byte': (bytes, bytes),
    'nonNegativeInteger': (int, 'numerical'),
    'unsignedLong': (int, 'numerical'),
    'unsignedInt': (int, 'numerical'),
    'unsignedShort': (int, 'numerical'),
    'positiveInteger': (int, 'numerical')
}

numerical = (int, float)

class Xbind(object):

    name = "BIND"

    def __init__(self, bind):
        self.input = Queue()
        self.qresults = Queue()
        self.bind = bind

    def execute(self, left, dummy, out, processqueue=Queue()):
        self.left = left 
        self.qresults = out
        b_value = self.bind.term
        b_var = self.bind.new_var[1:]
        tuple = self.left.get(True)

        if isinstance(b_value, Argument) or isinstance(b_value, str):

            while (tuple != "EOF"):
                shouldInclude = self.BindTerm_str(tuple)
                if shouldInclude:
                    self.qresults.put(tuple)

                tuple = self.left.get(True)
        
        elif isinstance(b_value, Expression):

            while (tuple != "EOF"):
                res = self.BindTerm_expr(tuple, self.bind.term.op, self.bind.term.left,
                                                        self.bind.term.right)  
                tuple[b_var] = res

                if res:
                    self.qresults.put(tuple)
                   
                tuple = self.left.get(True)

        # Put EOF in queue and exit.
        self.qresults.put("EOF")


    def BindTerm_str(self, tuple):
        isValid = True
        b_var = self.bind.new_var[1:]
        term = self.bind.term
        if isinstance(term, Argument) and (term.name[0] == "\"" or term.name[0] == "\'"):
            b_value = term.name[1:-1]
        else:
            b_value = term
        tuple[b_var] = b_value
        value = tuple[b_var]

        if value != b_value:
            isValid = False

        if isValid:
            return True

        return False          


    def BindTerm_expr(self, tuple, operator, term_left, term_right):
        
        # case 1: Var op NUMBER | Var op Var 
        if isinstance(term_left, Argument) and isinstance(term_right, Argument):
            if term_left.constant:
                res_left = term_left.name
            elif term_left.name.isnumeric() or term_left.name.replace('.','',1).isdigit():
                res_left = term_left.name
            else:
                res_left = self.extractValue(tuple[term_left.name[1:]])
            
            if term_right.constant:
                res_right = term_right.name
            elif term_right.name.isnumeric() or term_right.name.replace('.','',1).isdigit():
                res_right = term_right.name
            else:
                res_right = tuple[term_right.name[1:]]
            
            res = self.evaluateOperator(operator, res_left, res_right)

        # case 2: VAR op Expr | NUMBER op Expr
        elif isinstance(term_left, Argument) and isinstance(term_right, Expression):
   
            if term_left.constant: 
                res_left = term_left.name
            elif term_left.name.isnumeric() or term_left.name.replace('.','',1).isdigit():
                res_left = term_left.name
            else:
                res_left = tuple[term_left.name[1:]]

            res_right = self.extractExpr(tuple, term_right)
            res = self.evaluateOperator(operator, res_left, res_right)

        # case 3: Expr op VAR | Expr op NUMBER
        elif isinstance(term_right, Argument) and isinstance(term_left, Expression):

            if term_right.constant: 
                res_right = term_right.name
            elif term_right.name.isnumeric() or term_right.name.replace('.','',1).isdigit():
                res_right = term_right.name
            else:
                res_right = tuple[term_right.name[1:]]

            if term_left.right is not None:
                res_left = self.extractExpr(tuple, term_left)
                res = self.evaluateOperator(operator, res_left, res_right)
            
            if term_left.right is None:
                left = tuple[term_left.left.name[1:]]
                res_left = self.evaluateUnaryOperator(term_left.op, left)
                res = self.evaluateOperator(operator, res_left, res_right)            
            
        # case 4: unary_operators
        if isinstance(term_left, Argument) and term_right is None:

            if term_left.constant or term_left.name[1:] not in tuple: 
                res_left = term_left.name 
            else:
                res_left = tuple[term_left.name[1:]]
            
            res = self.evaluateUnaryOperator(operator, res_left)

        if isinstance(term_left, Expression) and term_right is None:

            if term_left.constant or term_left.name[1:] not in tuple: 
                res_left = term_left.name 
            else:
                res_left = tuple[term_left.name[1:]]
            
            res = self.evaluateUnaryOperator(operator, res_left)

        # case 5: logical connectives 
        if isinstance(term_left, Expression) and isinstance(term_right, Expression):

            if term_left.right is not None and term_right.right is not None:
                res_left = self.extractExpr(tuple, term_left)
                res_right = self.extractExpr(tuple, term_right)
                res = self.evaluateOperator(operator, res_left, res_right)

            elif term_left.right is None and term_right.right is None:
                left = tuple[term_left.left.name[1:]]
                right = tuple[term_right.left.name[1:]]
                res_left = self.evaluateUnaryOperator(term_left.op, left)
                res_right = self.evaluateUnaryOperator(term_right.op, right)
                res = self.evaluateOperator(operator, res_left, res_right)

        return str(res)


    # For nested_term:
    def extractExpr(self, tuple, nested_term):
        # right_term is Expr
        if isinstance(self.bind.term.right, Expression):
            nested_term_r = self.bind.term.right.right
            nested_term_l = self.bind.term.right.left
            nested_term_o = self.bind.term.right.op
        
        # left_term is Expr
        elif isinstance(self.bind.term.left, Expression):
            nested_term_r = self.bind.term.left.right
            nested_term_l = self.bind.term.left.left
            nested_term_o = self.bind.term.left.op


        # Expr: VAR op (NUMBER op Var) | VAR op (VAR op NUMBER) | FLOAT op (NUMBER op VAR)
        if isinstance(nested_term_l, Argument) and isinstance(nested_term_r, Argument):
            if str(nested_term_l).isnumeric() or str(nested_term_l).replace('.','',1).isdigit():
                if nested_term_r.constant:
                    res_right = nested_term_r.name
                else:
                    res_right = tuple[nested_term_r.name[1:]]
                
                res = self.evaluateOperator(nested_term_o, nested_term_l, res_right)
            
            elif str(nested_term_r).isnumeric() or str(nested_term_r).replace('.','',1).isdigit():
                if nested_term_l.constant:
                    res_left = nested_term_l.name
                else:
                    res_left = tuple[nested_term_l.name[1:]]
                
                res = self.evaluateOperator(nested_term_o, res_left, nested_term_r)

        # Expr: NUMBER op (NUMBER op VAR) 
        if isinstance(nested_term_l, str) and isinstance(nested_term_r, Argument):
            if nested_term_r.constant:
                res_right = nested_term_r.name
            else:
                res_right = tuple[nested_term_r.name[1:]]

            res = self.evaluateOperator(nested_term_o, nested_term_l, res_right)

        # Expr: NUMBER op (VAR op NUMBER)
        if isinstance(nested_term_l, Argument) and isinstance(nested_term_r, str):
            if nested_term_l.constant:
                res_left = nested_term_l.name
            else:
                res_left = tuple[nested_term_l.name[1:]] 

            res = self.evaluateOperator(nested_term_o, res_left, nested_term_r)           
            
        return res

    
    def evaluateOperator(self, operator, term_left, term_right):
        if operator in arithmetic_operators:
            return self.evaluateAritmethic(operator, term_left, term_right)
        elif (operator in unary_operators):
            return self.evaluateUnaryOperator(operator, term_left)
        elif (operator in test_operators):
            return self.evaluateTest(operator, term_left, term_right)
        elif (operator in logical_connectives):
            return self.evaluateLogicalConnective(operator, term_left, term_right)
            

    def evaluateAritmethic(self, operator, term_left, term_right):
        term_left = str(term_left)
        term_right = str(term_right)

        if term_left.isdigit() and term_right.isdigit():
            term_left = int(term_left)
            term_right = int(term_right)
            return (arithmetic_operators[operator](term_left, term_right))

        elif term_right.replace('.','',1).isdigit() and term_left.isdigit():
            term_right = float(term_right)
            term_left = int(term_left)
            return (arithmetic_operators[operator](term_left, term_right))

        elif term_left.replace('.','',1).isdigit() and term_right.isdigit():
            term_left = float(term_left)
            term_right = int(term_right)
            return (arithmetic_operators[operator](term_left, term_right))
        
        elif term_left.replace('.','',1).isdigit() and term_right.replace('.','',1).isdigit():
            term_left = float(term_left)
            term_right = float(term_right)
            return (arithmetic_operators[operator](term_left, term_right))

        else:
            raise SPARQLTypeError


    def evaluateUnaryOperator(self, operator, term_left):

        if (operator.lower() == 'floor'):
            return (unary_operators[operator.lower()](math.floor(float(term_left))),int)

        elif (operator.lower() == 'str'):
            return (unary_operators[operator.lower()](str(term_left)), str)

        elif (operator.lower() == 'bound'):
            if term_left[0] != "?":
                return (unary_operators[operator.lower()](term_left), bool)
            elif term_left[0] == "?":
                return (False, bool)

        elif (operator.lower() == 'year'):
            
            for format in ("%Y-%m-%dT%H:%M:%SZ", "%Y-%m-%d %H:%M:%SZ", "%Y-%m-%d %H:%M:%S UTC", "%Y-%m-%d"):
                try:
                    term_left = str(term_left)
                    date = datetime.datetime.strptime(term_left, format)
                    return int(date.year)
                    
                except ValueError:
                    pass
                
            raise ValueError("no valid date format found")

        else:
            raise SPARQLTypeError


    def evaluateTest(self, operator, left, right):

        if left and right:
            (term_left, type_left) = left, type(left)
            (term_right, type_right) = right, type(right)
            term_left = str(term_left)
            term_right = str(term_right)
        else:
            print("Exception: left or right operand None.", left, right)
            raise Exception

        if term_left.isnumeric() and term_right.isnumeric():
            term_left = int(term_left)
            term_right = int(term_right)
            return (test_operators[operator](term_left, term_right), bool)

        elif term_left.replace('.','',1).isdigit() and term_right.isnumeric():
            term_left = float(term_left)
            term_right = int(term_right)
            return (test_operators[operator](term_left, term_right), bool)

        elif term_left.isnumeric() and term_right.replace('.','',1).isdigit():
            term_left = int(term_left)
            term_right = float(term_right)
            return (test_operators[operator](term_left, term_right), bool)

        elif term_left.replace('.','',1).isdigit() and term_right.replace('.','',1).isdigit():
            term_left = float(term_left)
            term_right = float(term_right)
            return (test_operators[operator](term_left, term_right), bool)

        else:
            print("SPARQLTypeError - in Xbind")
            raise SPARQLTypeError


    def evaluateLogicalConnective(self, operator, term_left, term_right):
        
        (isEBV_left, ebv_left) = self.evaluateEBV(term_left)
        (isEBV_right, ebv_right) = self.evaluateEBV(term_right)

        if (isEBV_left and isEBV_right):
            return (logical_connectives[operator](ebv_left, ebv_right), bool)

        elif (isEBV_left):
            res = logical_connectives[operator](ebv_left, 'Error')
            if (res == 'Error'):
                raise SPARQLTypeError
            else:
                return (res, bool)

        elif (isEBV_right):
            res = logical_connectives[operator](ebv_right, 'Error')
            if (res == 'Error'):
                raise SPARQLTypeError
            else:
                return (res, bool)

    '''
    evaluateEBV: calculates whether an argument is an Effective Boolean Value (EBV)
                 according to the definition in the SPARQL documentation 
                 See: http://www.w3.org/TR/sparql11-query/#ebv

    input: val -- an argument
    return: (isEBV, EBV) -- both of Python type bool
    you still need to make it consistent with XFilter

    '''

    def evaluateEBV(self, value):
        (casted_val, type_val) = value

        # Handles python data types.
        if (isinstance(casted_val, bool)):
            return (True, casted_val)

        # Rule 1
        if ((type_val == bool) and (casted_val != 'True') and (casted_val != 'False')):
            return (True, False)

        # Rule 2
        if (type_val == bool):
            if (casted_val == 'True'):
                return (True, True)
            elif (casted_val == 'False'):
                return (True, False)

        # Rule 3: !Bound
        if (type_val == str):
            if casted_val[0] == "?":
                return (True, False)
            elif casted_val.isnumeric():
                return (True, True)

        # Rule 4: The error type should be raised by the evaluators.
        return (False, None)


    def extractValue(self, val):
        pos = val.find("^^")
        # Handles when the literal is typed.
        if (pos > -1):
            for t in data_types.keys():
                if (t in val[pos:]):    
                    (python_type, general_type) = data_types[t]
                    if (general_type == bool):
                        return val[:pos]
                    else:
                        return python_type(val[:pos])
        else:
            return str(val)


class SPARQLTypeError(Exception):
    """Base class for exceptions in this module."""
    pass


"""
Created on 23/10/2022.

Implements the temporary handling of Expression. 
Note that this handle only Arithmetic expression for now.

@author: Avellino
"""

from DeTrusty.Sparql.Parser.services import Argument, Aggregate, Expression
import operator

arithmetic_operators = {
    '*': operator.mul,
    '/': operator.truediv,
    '+': operator.add,
    '-': operator.sub
}

def simplifyExp(exp, tuple):
    if type(exp) is Argument:
        if exp.constant: 
            if exp.name == '*':
                for el in tuple:
                    if type(el) is list:
                        return el
            return exp.name
        else:
            return tuple[exp.name[1:]]
    else:
        return evaluateOperator(exp, tuple)

def evaluateOperator(exp, tuple):

    if type(exp) is Aggregate:
        if type(exp.exp) is Expression or type(exp.exp) is Argument:
            ls = simplifyExp(exp.exp, tuple)
        return evaluateAggregate(exp.op, exp.distinct, exp.sep, ls)

    if type(exp.left) is Expression or type(exp.left) is Argument or type(exp.left) is Aggregate:
        simp_left = simplifyExp(exp.left, tuple)

    if type(exp.right) is Expression or type(exp.right) is Argument or type(exp.right) is Aggregate: 
        simp_right = simplifyExp(exp.right, tuple)

    if exp.exp_type == 'arithmetic':
        if type(simp_left) is list and type(simp_right) is list:
            ret = list()
            for el1 in simp_left:
                for el2 in simp_right:
                    ret.append(evaluateArithmetic(el1, exp.op, el2))
            return ret
        if type(simp_left) is list:
            ret = list()
            for el in simp_left:
                ret.append(evaluateArithmetic(el, exp.op, simp_right))
            return ret
        if type(simp_right) is list:
            ret = list()
            for el in simp_right:
                ret.append(evaluateArithmetic(simp_left, exp.op, el))
            return ret
        return evaluateArithmetic(simp_left, exp.op, simp_right)

def evaluateArithmetic(left, op, right):
    try:
        return arithmetic_operators[op](float(left), float(right))
    except:
        return ''

# distinction of RDF literal: https://www.w3.org/TR/rdf-concepts/#section-Literal-Equality
def evaluateAggregate(op, dist, sep, ls):

    if op.upper() == 'COUNT':
        if dist:
            rec = list()
            count = 0
            for val in ls:
                if val not in rec and val != '': 
                    rec.append(val)
                    count += 1
            return count

        else:
            count = 0
            for val in ls:
                if val != '':
                    count += 1
            return count

    if op.upper() == 'GROUP_CONCAT':
        if dist:
            tmp = set(ls)
            if '' in tmp:
                tmp.remove('')
            if len(tmp) > 0:
                tmp1 = str(tmp)
                if sep:
                    tmp2 = tmp1.replace("'", "")
                    ret = tmp2.replace(", ", sep[1:len(sep)-1])
                else:
                    ret = tmp1.replace("'", "")
                return ret[1:len(ret)-1]
            else:
                return ''
        else:
            tmp = ls.copy()
            while '' in tmp:
                tmp.remove('')
            if len(tmp) > 0:
                tmp1 = str(tmp)
                if sep:
                    tmp2 = tmp1.replace("'", "")
                    ret = tmp2.replace(", ", sep[1:len(sep)-1])
                else:
                    ret = tmp1.replace("'", "")
                return ret[1:len(ret)-1]
            else:
                return ''

    if op.upper() == 'SAMPLE':
        for val in ls:
            if val != '':
                return val
        return ''

    if op.upper() == 'MIN':
        tmp = None
        if dist:
            tmp = set(ls)
            if '' in tmp:
                tmp.remove('')
            tmp = list(tmp)
        else:
            tmp = ls.copy()
            while '' in tmp:
                tmp.remove('')
        tmp.sort()
        if tmp:
            return tmp[0]
        else:
            return ''

    if op.upper() == 'MAX':
        tmp = None
        if dist:
            tmp = set(ls)
            if '' in tmp:
                tmp.remove('')
            tmp = list(tmp)
        else:
            tmp = ls.copy()
            while '' in tmp:
                tmp.remove('')
        tmp.sort(reverse=True)
        if tmp:
            return tmp[0]
        else:
            return ''

    if op.upper() == 'SUM':
        ret = 0
        if dist:
            tmp = list()
            for val in ls:
                if val != '' and val not in tmp:
                    tmp.append(val)
                    try:
                        ret += float(val)
                    except:
                        return ''
        else:
            for val in ls:
                if val != '':
                    try:
                        ret += float(val)
                    except:
                        return ''
        return ret

    if op.upper() == 'AVG':
        ret = 0
        val_count = 0
        if dist:
            tmp = list()
            for val in ls:
                if val != '' and val not in tmp:
                    tmp.append(val)
                    try:
                        ret += float(val)
                    except:
                        return ''
                    val_count += 1
        else:
            for val in ls:
                if val != '':
                    try:
                        ret += float(val)
                    except:
                        return ''
                    val_count += 1
        if val_count != 0:
            ret = ret / val_count
        return ret



import urllib.parse
from ply import lex, yacc
from .services import Query, Argument, Triple, UnionBlock, JoinBlock, Optional, Filter, Expression, Service

# Lexer
reserved = {
    'UNION': 'UNION',
    'FILTER': 'FILTER',
    'OPTIONAL': 'OPTIONAL',
    'SELECT': 'SELECT',
    'DISTINCT': 'DISTINCT',
    'WHERE': 'WHERE',
    'PREFIX': 'PREFIX',
    'SERVICE': 'SERVICE',
    'LIMIT': 'LIMIT',
    'OFFSET': 'OFFSET',
    'ORDER': 'ORDER',
    'BY': 'BY',
    'DESC': 'DESC',
    'ASC': 'ASC',
    'BOUND': 'BOUND',
    'REGEX': 'REGEX',
    'ISIRI': 'ISIRI',
    'ISURI': 'ISURI',
    'ISBLANK': 'ISBLANK',
    'ISLITERAL': 'ISLITERAL',
    'LANG': 'LANG',
    'DATATYPE': 'DATATYPE',
    'SAMETERM': 'SAMETERM',
    'LANGMATCHES': 'LANGMATCHES',
    'STR': 'STR',
    'UPPERCASE': 'UPPERCASE'
}

tokens = [
    "CONSTANT",
    "VARIABLE",
    "NUMBER",
    "LKEY",
    "RKEY",
    "COLON",
    "POINT",
    "URI",
    "ALL",
    "LPAR",
    "RPAR",
    "EQUALS",
    "LESS",
    "LESSEQ",
    "GREATER",
    "GREATEREQ",
    "ID",
    "COMA",
    "NEQUALS",
    "NEG",
    "AND",
    "PLUS",
    "MINUS",
    "TIMES",
    "DIV",
    "DOUBLE",
    "INTEGER",
    "DECIMAL",
    "FLOAT",
    "STRING",
    "BOOLEAN",
    "DATETIME",
    "NONPOSINT",
    "NEGATIVEINT",
    "LONG",
    "INT",
    "SHORT",
    "BYTE",
    "NONNEGINT",
    "UNSIGNEDLONG",
    "UNSIGNEDINT",
    "UNSIGNEDSHORT",
    "UNSIGNEDBYTE",
    "POSITIVEINT",
    "OR"
] + list(reserved.values())


def t_ID(t):
    r'[a-zA-Z_][a-zA-Z_0-9\-]*'
    t.type = reserved.get(t.value.upper(), 'ID')  # Check for reserved words
    return t


t_CONSTANT = r"(\"|\')[^\"\'\n\r]*(\"|\')((@[a-z][a-z]) | (\^\^\w+))?"
# t_CONSTANT = r"(\"|\')[^\"\'\n\r]*(\"|\')(@[a-z][a-z])?" # According to ISO 639-1, lang tags are specified with two letters.
# t_CONSTANT = r"(\"|\')[^\"\'\n\r]*(\"|\')(@en)?"
# t_CONSTANT = r"(\"|\')[^\"\'\n\r]*(\"|\')"
t_NUMBER = r"([0-9])+"
t_VARIABLE = r"([\?]|[\$])([A-Z]|[a-z])\w*"
t_LKEY = r"\{"
t_LPAR = r"\("
t_RPAR = r"\)"
t_COLON = r"\:"
t_ALL = r"\*"
t_RKEY = r"(\.)?\s*\}"
t_POINT = r"\."
t_LESS = r"<"
t_LESSEQ = r"<="
t_GREATER = r">"
t_GREATEREQ = r">="
t_URI = r"<\S+>"
t_COMA = r"\,"
t_EQUALS = r"="
t_NEQUALS = r"\!="
t_NEG = r"\!"
t_AND = r"\&\&"
t_OR = r"\|\|"
t_PLUS = r"\+"
t_MINUS = r"\-"
t_TIMES = r"\*"
t_DIV = r"/"
t_DOUBLE = r"xsd\:double"
t_INTEGER = r"xsd\:integer"
t_DECIMAL = r"xsd\:decimal"
t_FLOAT = r"xsd\:float"
t_STRING = r"xsd\:string"
t_BOOLEAN = r"xsd\:boolean"
t_DATETIME = r"xsd\:dateTime"
t_NONPOSINT = r"xsd\:nonPositiveInteger"
t_NEGATIVEINT = r"xsd\:negativeInteger"
t_LONG = r"xsd\:long"
t_INT = r"xsd\:int"
t_SHORT = r"xsd\:short"
t_BYTE = r"xsd\:byte"
t_NONNEGINT = r"xsd\:nonNegativeInteger"
t_UNSIGNEDLONG = r"xsd\:unsignedLong"
t_UNSIGNEDINT = r"xsd\:unsignedInt"
t_UNSIGNEDSHORT = r"xsd\:unsignedShort"
t_UNSIGNEDBYTE = r"xsd\:unsignedByte"
t_POSITIVEINT = r"xsd\:positiveInteger"
t_ignore = ' \t\n'


def t_error(t):
    raise TypeError("Unknown text '%s' in line %d " % (t.value, t.lexer.lineno,))


# Define a rule so we can track line numbers
def t_newline(t):
    r'\n+'
    t.lexer.lineno += len(t.value)


lexer = lex.lex()


# Parser
def p_parse_sparql(p):
    """
    parse_sparql : prefix_list query order_by limit offset
    """
    (vs, ts, d) = p[2]
    p[0] = Query(prefs=p[1], args=vs, body=ts, distinct=d, order_by=p[3], limit=p[4], offset=p[5])


def p_prefix_list(p):
    """
    prefix_list : prefix prefix_list
    """
    p[0] = [p[1]] + p[2]


def p_empty_prefix_list(p):
    """
    prefix_list : empty
    """
    p[0] = []


def p_empty(p):
    """
    empty :
    """
    pass


def p_prefix(p):
    """
    prefix : PREFIX uri
    """
    p[0] = p[2]


def p_uri_0(p):
    """
    uri : ID COLON ID
    """
    p[0] = p[1] + p[2] + p[3]


def p_uri_1(p):
    """
    uri : ID COLON URI
    """
    p[0] = p[1] + p[2] + p[3]


def p_uri_2(p):
    """
    uri : URI
    """
    p[0] = p[1]


def p_order_by_0(p):
    """
    order_by : ORDER BY var_order_list desc_var
    """
    p[0] = p[3] + [p[4]]


def p_order_by_1(p):
    """
    order_by : empty
    """
    p[0] = []


def p_var_order_list_0(p):
    """
    var_order_list : empty
    """
    p[0] = []


def p_var_order_list_1(p):
    """
    var_order_list : var_order_list desc_var
    """
    p[0] = p[1] + [p[2]]


def p_desc_var_0(p):
    """
    desc_var : DESC LPAR VARIABLE RPAR
    """
    p[0] = Argument(p[3], False, True)


def p_desc_var_1(p):
    """
    desc_var : VARIABLE
    """
    p[0] = Argument(p[1], False, False)


def p_desc_var_2(p):
    """
    desc_var : ASC LPAR VARIABLE RPAR
    """
    p[0] = Argument(p[3], False, False)


def p_desc_var_3(p):
    """
    desc_var : unary_func LPAR desc_var RPAR
    """
    p[0] = Expression(p[1], p[3], None)


def p_limit_0(p):
    """
    limit : LIMIT NUMBER
    """
    p[0] = p[2]


def p_limit_1(p):
    """
    limit : empty
    """
    p[0] = -1


def p_offset_0(p):
    """
    offset : OFFSET NUMBER
    """
    p[0] = p[2]


def p_offset_1(p):
    """
    offset : empty
    """
    p[0] = -1


def p_query_0(p):
    """
    query : SELECT distinct var_list WHERE LKEY group_graph_pattern RKEY
    """
    p[0] = (p[3], p[6], p[2])


def p_query_1(p):
    """
    query : SELECT distinct ALL WHERE LKEY group_graph_pattern RKEY
    """
    p[0] = ([], p[6], p[2])


def p_distinct_0(p):
    """
    distinct : DISTINCT
    """
    p[0] = True


def p_distinct_1(p):
    """
    distinct : empty
    """
    p[0] = False


# def p_service_list(p):
#    """
#    service_list : service_list POINT service
#    """
#    p[0] = p[1] + [p[3]]


# def p_single_service_list(p):
#    """
#    service_list : service
#    """
#    p[0] = [p[1]]


def p_service(p):
    """
    service : SERVICE uri LKEY group_graph_pattern_service RKEY
    """
    p[0] = Service(p[2], p[4])


def p_ggp_service_0(p):
    """
    group_graph_pattern_service : union_block_service
    """
    p[0] = UnionBlock(p[1])


def p_union_block_service_0(p):
    """
    union_block_service : join_block_service rest_union_block_service
    """
    p[0] = [JoinBlock(p[1])] + p[2]


def p_rest_union_block_service_0(p):
    """
    rest_union_block_service : empty
    """
    p[0] = []


def p_rest_union_block_service_1(p):
    """
    rest_union_block_service : UNION LKEY join_block_service rest_union_block_service RKEY rest_union_block_service
    """
    p[0] = [JoinBlock(p[3])] + p[4] + p[6]


def p_join_block_service_0(p):
    """
    join_block_service : LKEY bgp_service rest_join_block_service RKEY rest_join_block_service
    """
    jb_list = [p[2]] + p[3]
    if p[5] != [] and isinstance(p[5][0], Filter):
        p[0] = [UnionBlock([JoinBlock(jb_list)])] + p[5]
    elif isinstance(p[2], UnionBlock):
        p[0] = [p[2]] + p[3] + p[5]
    else:
        p[0] = [UnionBlock([JoinBlock(jb_list)])] + p[5]


def p_join_block_service_1(p):
    """
    join_block_service : bgp_service rest_join_block_service
    """
    p[0] = [p[1]] + p[2]


def p_rest_join_block_service_0(p):
    """
    rest_join_block_service : empty
    """
    p[0] = []


def p_rest_join_block_service_1(p):
    """
    rest_join_block_service : POINT bgp_service rest_join_block_service
    """
    p[0] = [p[2]] + p[3]


def p_rest_join_block_service_2(p):
    """
    rest_join_block_service : bgp_service rest_join_block_service
    """
    p[0] = [p[1]] + p[2]


def p_bgp_service_01(p):
    """
    bgp_service :  LKEY join_block_service UNION join_block_service rest_union_block_service RKEY
    """
    ggp = [JoinBlock(p[2])] + [JoinBlock(p[4])] + p[5]
    p[0] = UnionBlock(ggp)


def p_bgp_service_02(p):
    """
    bgp_service :  join_block_service UNION join_block_service rest_union_block_service
    """
    ggp = [JoinBlock(p[1])] + [JoinBlock(p[3])] + p[4]
    p[0] = UnionBlock(ggp)


def p_bgp_service_1(p):
    """
    bgp_service : triple
    """
    p[0] = p[1]


def p_bgp_service_2(p):
    """
    bgp_service : FILTER LPAR expression RPAR
    """
    p[0] = Filter(p[3])


def p_bgp_service_3(p):
    """
    bgp_service : FILTER express_rel
    """
    p[0] = Filter(p[2])


def p_bgp_service_4(p):
    """
    bgp_service : OPTIONAL LKEY group_graph_pattern_service RKEY
    """
    p[0] = Optional(p[3])


def p_bgp_service_5(p):
    """
    bgp_service : LKEY join_block_service rest_union_block_service RKEY
    """
    bgp_arg = p[2] + p[3]
    p[0] = UnionBlock(JoinBlock(bgp_arg))


def p_var_list(p):
    """
    var_list : var_list VARIABLE
    """
    p[0] = p[1] + [Argument(p[2], False)]


def p_single_var_list(p):
    """
    var_list : VARIABLE
    """
    p[0] = [Argument(p[1], False)]


def p_ggp_0(p):
    """
    group_graph_pattern : union_block
    """
    p[0] = UnionBlock(p[1])


def p_union_block_0(p):
    """
    union_block : join_block rest_union_block
    """
    (jb, filters) = p[1]
    p[0] = [JoinBlock(jb, filters)] + p[2]


def p_rest_union_block_0(p):
    """
    rest_union_block : empty
    """
    p[0] = []


def p_rest_union_block_1(p):
    """
    rest_union_block : UNION LKEY join_block rest_union_block RKEY rest_union_block
    """
    (jb, filters) = p[3]
    p[0] = [JoinBlock(jb, filters)] + p[4] + p[6]


def p_join_block(p):
    """
    join_block : bgp rest_join_block
    """
    (jb2, filters2) = p[2]
    (jb1, filters1) = p[1]
    p[0] = (jb1 + jb2, filters1 + filters2)


def p_rest_join_block_0(p):
    """
    rest_join_block : empty
    """
    p[0] = ([], [])


def p_rest_join_block_1(p):
    """
    rest_join_block : POINT bgp rest_join_block
    """
    (jb3, filters3) = p[3]
    (jb2, filters2) = p[2]
    p[0] = (jb2 + jb3, filters2 + filters3)


def p_rest_join_block_2(p):
    """
    rest_join_block : bgp rest_join_block
    """
    (jb2, filters2) = p[2]
    (jb1, filters1) = p[1]
    p[0] = (jb1 + jb2, filters1 + filters2)


def p_bgp_00(p):
    """
    bgp : LKEY bgp  UNION bgp rest_union_block  RKEY
    """
    (jb1, filters1) = p[2]
    (jb3, filters3) = p[4]
    ggp = [JoinBlock(jb1, filters1)] + [JoinBlock(jb3, filters3)] + p[5]
    ggp1 = [UnionBlock(ggp)]
    p[0] = (ggp1, [])


def p_bgp_01(p):
    """
    bgp : bgp  UNION bgp rest_union_block
    """
    (jb1, filters1) = p[1]
    (jb3, filters3) = p[3]
    ggp = [JoinBlock(jb1, filters1)] + [JoinBlock(jb3, filters3)] + p[4]
    ggp1 = [UnionBlock(ggp)]
    p[0] = (ggp1, [])


# def p_bgp_01(p):
#    """
#    bgp : join_block  UNION join_block rest_union_block
#    """
#    (jb1,filters1) = p[1]
#    (jb3,filters3) = p[3]
#    ggp = [JoinBlock(jb1,filters1)] + [JoinBlock(jb3,filters3)] + p[4]
#    ggp1 = [UnionBlock(ggp)]
#    p[0]=(ggp1,[])


def p_bgp_11(p):
    """
    bgp : service
    """
    p[0] = ([p[1]], [])


def p_bgp_1(p):
    """
    bgp : LKEY service RKEY
    """
    p[0] = ([p[2]], [])


# def p_bgp_1(p):
#    """
#    bgp : service
#    """
#    p[0] = ([p[1]],[])


def p_bgp_2(p):
    """
    bgp : FILTER LPAR expression RPAR
    """
    p[0] = ([], [Filter(p[3])])


def p_bgp_3(p):
    """
    bgp : FILTER express_rel
    """
    p[0] = ([], [Filter(p[2])])


def p_bgp_4(p):
    """
    bgp : triple
    """
    p[0] = ([p[1]], [])


def p_bgp_5(p):
    """
    bgp : OPTIONAL LKEY group_graph_pattern RKEY
    """
    p[0] = ([Optional(p[3])], [])


def p_bgp_6(p):
    """
    bgp : LKEY join_block RKEY
    """
    (jb, filters) = p[2]
    if (len(jb) == 1):
        p[0] = (jb, filters)
    else:
        p[0] = ([JoinBlock(jb, filters)], [])


def p_triple_0(p):
    """
    triple : subject predicate object
    """
    p[0] = Triple(p[1], p[2], p[3])


def p_expression_0(p):
    """
    expression : express_rel LOGOP expression
    """
    p[0] = Expression(p[2], p[1], p[3])


def p_expression_1(p):
    """
    expression : express_rel
    """
    p[0] = p[1]


def p_expression_2(p):
    """
    expression : LPAR expression RPAR
    """
    p[0] = p[2]


def p_express_rel_0(p):
    """
    express_rel : express_arg RELOP express_rel
    """
    p[0] = Expression(p[2], p[1], p[3])


def p_express_rel_1(p):
    """
    express_rel : express_arg
    """
    p[0] = p[1]


def p_express_rel_2(p):
    """
    express_rel : LPAR express_rel RPAR
    """
    p[0] = p[2]


def p_express_rel_3(p):
    """
    express_rel : NEG LPAR expression RPAR
    """
    p[0] = Expression(p[1], p[3], None)


def p_express_rel_4(p):
    """
    express_rel : NEG express_rel
    """
    p[0] = Expression(p[1], p[2], None)


def p_express_arg_0(p):
    """
    express_arg : uri
    """
    p[0] = Argument(p[1], True)


def p_express_arg_1(p):
    """
    express_arg : VARIABLE
    """
    p[0] = Argument(p[1], False)


def p_express_arg_2(p):
    """
    express_arg : CONSTANT
    """
    p[0] = Argument(p[1], True)


def p_express_arg_3(p):
    """
    express_arg : NUMBER
    """
    p[0] = Argument(p[1], True)


def p_express_arg_03(p):
    """
    express_arg : NUMBER POINT NUMBER
    """
    numberDecimal = str(p[1]) + p[2] + str(p[3])
    p[0] = Argument(numberDecimal, True)


def p_express_arg_4(p):
    """
    express_arg : REGEX LPAR express_arg COMA pattern_arg regex_flag
    """
    p[0] = Expression("REGEX", p[3], Argument(p[5], False, p[6]))


def p_regex_flags_0(p):
    """
    regex_flag : RPAR
    """
    p[0] = False


def p_regex_flags_1(p):
    """
    regex_flag : COMA pattern_arg RPAR
    """
    p[0] = p[2]


def p_pattern_arg_0(p):
    """
    pattern_arg : CONSTANT
    """
    p[0] = p[1]


def p_express_arg_5(p):
    """
    express_arg : binary_func LPAR express_arg COMA express_arg RPAR
    """
    p[0] = Expression(p[1], p[3], p[5])


def p_express_arg_6(p):
    """
    express_arg : unary_func LPAR express_arg RPAR
    """
    p[0] = Expression(p[1], p[3], None)


def p_express_arg_7(p):
    """
    express_arg : UNARYOP express_arg
    """
    p[0] = Expression(p[1], p[2], None)


def p_express_arg_8(p):
    """
    express_arg : express_arg ARITOP express_arg
    """
    p[0] = Expression(p[2], p[1], p[3])


def p_express_arg_9(p):
    """
    express_arg : LPAR express_arg RPAR
    """
    p[0] = p[2]


def p_express_arg_10(p):
    """
    express_arg : express_arg RELOP express_arg
    """
    p[0] = Expression(p[2], p[1], p[3])


def p_arit_op_0(p):
    """
    ARITOP : PLUS
    """
    p[0] = p[1]


def p_arit_op_1(p):
    """
    ARITOP : MINUS
    """
    p[0] = p[1]


def p_arit_op_2(p):
    """
    ARITOP : TIMES
    """
    p[0] = p[1]


def p_arit_op_3(p):
    """
    ARITOP : DIV
    """
    p[0] = p[1]


def p_unaryarit_op_1(p):
    """
    UNARYOP : PLUS
    """
    p[0] = p[1]


def p_unaryarit_op_2(p):
    """
    UNARYOP : MINUS
    """
    p[0] = p[1]


def p_logical_op_0(p):
    """
    LOGOP : AND
    """
    p[0] = p[1]


def p_logical_op_1(p):
    """
    LOGOP : OR
    """
    p[0] = p[1]


def p_relational_op_0(p):
    """
    RELOP : EQUALS
    """
    p[0] = p[1]


def p_relational_op_1(p):
    """
    RELOP : LESS
    """
    p[0] = p[1]


def p_relational_op_2(p):
    """
    RELOP : LESSEQ
    """
    p[0] = p[1]


def p_relational_op_3(p):
    """
    RELOP : GREATER
    """
    p[0] = p[1]


def p_relational_op_4(p):
    """
    RELOP : GREATEREQ
    """
    p[0] = p[1]


def p_relational_op_5(p):
    """
    RELOP : NEQUALS
    """
    p[0] = p[1]


def p_binary_0(p):
    """
    binary_func : REGEX
    """
    p[0] = "REGEX"


def p_binary_1(p):
    """
    binary_func : SAMETERM
    """
    p[0] = "sameTERM"


def p_binary_2(p):
    """
    binary_func : LANGMATCHES
    """
    p[0] = "langMATCHES"


def p_unary_0(p):
    """
    unary_func : BOUND
    """
    p[0] = "BOUND"


def p_unary_1(p):
    """
    unary_func : ISIRI
    """
    p[0] = "ISIRI"


def p_unary_2(p):
    """
    unary_func : ISURI
    """
    p[0] = "ISURI"


def p_unary_3(p):
    """
    unary_func : ISBLANK
    """
    p[0] = "ISBLANK"


def p_unary_4(p):
    """
    unary_func : ISLITERAL
    """
    p[0] = "ISLITERAL"


def p_unary_5(p):
    """
    unary_func : LANG
    """
    p[0] = "LANG"


def p_unary_6(p):
    """
    unary_func : DATATYPE
    """
    p[0] = "DATATYPE"


def p_unary_7(p):
    """
    unary_func : STR
    """
    p[0] = "STR"


def p_unary_8(p):
    """
    unary_func : UPPERCASE
    """
    p[0] = "UPPERCASE"


def p_unary_9(p):
    """
    unary_func : DOUBLE
               | INTEGER
               | DECIMAL
               | FLOAT
               | STRING
               | BOOLEAN
               | DATETIME
               | NONPOSINT
               | NEGATIVEINT
               | LONG
               | INT
               | SHORT
               | BYTE
               | NONNEGINT
               | UNSIGNEDLONG
               | UNSIGNEDINT
               | UNSIGNEDSHORT
               | UNSIGNEDBYTE
               | POSITIVEINT
    """
    p[0] = p[1]


def p_unary_10(p):
    """
    unary_func : ID COLON ID
    """
    p[0] = p[1] + p[2] + p[3]


def p_unary_11(p):
    """
    unary_func : uri
    """
    p[0] = p[1]


def p_predicate_rdftype(p):
    """
    predicate : ID
    """
    if p[1] == 'a':
        value = '<http://www.w3.org/1999/02/22-rdf-syntax-ns#type>'
        # print('value set')
        p[0] = Argument(value, True)
    else:
        # print('raising')
        p_error(p[1])
        raise SyntaxError


def p_predicate_uri(p):
    """
    predicate : uri
    """
    p[0] = Argument(p[1], True)


def p_predicate_var(p):
    """
    predicate : VARIABLE
    """
    p[0] = Argument(p[1], False)


def p_subject_uri(p):
    """
    subject : uri
    """
    p[0] = Argument(p[1], True)


def p_subject_variable(p):
    """
    subject : VARIABLE
    """
    p[0] = Argument(p[1], False)


def p_object_uri(p):
    """
    object : uri
    """
    p[0] = Argument(p[1], True)


def p_object_variable(p):
    """
    object : VARIABLE
    """
    p[0] = Argument(p[1], False)


def p_object_constant(p):
    """
    object : CONSTANT
    """
    p[0] = Argument(p[1], True)


def p_error(p):
    raise TypeError("unknown text at %r" % (p.value,))


parser = yacc.yacc(debug=0, write_tables=False)


# Helpers
def parse(string):
    return parser.parse(urllib.parse.unquote(string), lexer=lexer)



__author__ = 'Kemele M. Endris and Philipp D. Rohde'

import collections
from enum import Enum
from functools import partial

from DeTrusty.Logger import get_logger
from DeTrusty.Decomposer import Tree, utils
from DeTrusty.Sparql.Parser.services import Service, Triple, Filter, Optional, UnionBlock, JoinBlock, Values, Bind, Argument

logger = get_logger(__name__, '.decompositions.log')
class Decomposer(object):

    def __init__(self, query, config, decompType="STAR", joinstarslocally=True, sparql_one_dot_one=False):
        self.decomposition_type = DecompositionType[decompType]
        self.sparql_one_dot_one = sparql_one_dot_one
        if not self.sparql_one_dot_one:
            from DeTrusty.Sparql.Parser import queryParser
            self.query = queryParser.parse(query)
        else:
            from DeTrusty.Sparql.Parser import queryParser1_1
            self.query = queryParser1_1.parse(query)
        if self.query:
            self.prefixes = utils.getPrefs(self.query.prefs)
        else:
            self.prefixes = None
        self.config = config
        self.joinlocally = joinstarslocally
        self.alltriplepatterns = []

    def decompose(self):

        if not self.query: # handling empty or whitespaces-filled queries, TODO: review error message from DeTrusty/__init__.py, it is not quite specific
            return None

        self.query.body = self.decomposeUnionBlock(self.query.body) if not self.sparql_one_dot_one else self.query.body

        if not self.query.body:
            return None

        logger.info('Decomposition obtained')
        logger.info(self.query)

        self.query.body = self.makePlanQuery(self.query)

        return self.query

    def decomposeUnionBlock(self, ub):
        r = []
        for jb in ub.triples:
            pjb = self.decomposeJoinBlock(jb)
            if pjb:
                r.append(pjb)
        if r:
            return UnionBlock(r)
        else:
            return None

    def decomposeJoinBlock(self, jb):
        """
        tl : triple
        sl : ???
        fl : filters, (additionally, BIND, VALUES)
        """
        tl = []
        sl = []
        fl = []
        for bgp in jb.triples:
            if isinstance(bgp, Triple):
                tl.append(bgp)
                self.alltriplepatterns.append(bgp)
            elif isinstance(bgp, Filter):
                fl.append(bgp)
            elif isinstance(bgp, Values):
                fl.append(bgp)
            elif isinstance(bgp, Bind):
                fl.append(bgp)
            elif isinstance(bgp, Optional):
                ubb = self.decomposeUnionBlock(bgp.bgg)
                skipp = False
                if ubb is not None:
                    for ot in ubb.triples:
                        if isinstance(ot, JoinBlock) and len(ot.triples) > 1 and len(ot.filters) > 0:
                            skipp = True
                            break
                    if not skipp:
                        sl.append(Optional(ubb))
            elif isinstance(bgp, UnionBlock):
                pub = self.decomposeUnionBlock(bgp)
                if pub:
                    sl.append(pub)
            elif isinstance(bgp, JoinBlock):
                pub = self.decomposeJoinBlock(bgp)
                if pub:
                    sl.append(pub)
        if tl:
            gs = self.decomposition_type.decompose(self, tl)
            if gs:
                gs.extend(sl)
                sl = gs
            else:
                return None

        fl1 = self.includeFilter(sl, fl)
        fl = list(set(fl) - set(fl1)) # in case VALUES with variables from multiple sources, the values object won't be removed
        if sl:
            if len(sl) == 1 and isinstance(sl[0], UnionBlock) and fl != []:
                sl[0] = self.updateFilters(sl[0], fl)
            j = JoinBlock(sl, filters=fl)
            return j
        else:
            return None

    def search(self, predicate: Argument) -> list:
        """
        Returns a list of all endpoints serving the specified predicate.
        """
        predicate = utils.getUri(predicate, self.prefixes)[1:-1]

        if predicate in self.config.predwrapidx:
            return list(self.config.predwrapidx[predicate])
        else:
            return []

    def decompose_triple_wise(self, tl):
        """
        Decomposes the query into triple-wise sub-queries.
        """
        views = []
        for sg in tl:
            eps = self.search(sg.predicate)
            if len(eps) == 1:
                serv = Service(eps[0], sg)
                views = views + [serv]
            if len(eps) > 1:
                elems = [JoinBlock([Service(ep, sg)]) for ep in eps]
                ub = UnionBlock(elems)
                views = views + [ub]
        return views

    def decompose_exclusive_groups(self, tl):
        """
        Decomposes the query into sub-queries following the exclusive groups approach.
        """
        qcl0 = collections.defaultdict(list)
        qcl1 = collections.defaultdict(list)

        for sg in tl:
            eps0 = self.search(sg.predicate)
            if not eps0:
                qcl0 = qcl1 = []
            elif len(eps0) == 1:
                p = eps0[0]
                qcl0[p].append(sg)
            else:
                qcl1[sg].extend(eps0)

        views0 = []
        views1 = []
        for cl in qcl0:
            l0 = qcl0[cl]
            serv = Service(cl, l0)
            views0 = views0 + [serv]
        for t in qcl1:
            eps = qcl1[t]
            elems = [JoinBlock([Service(ep, t)]) for ep in eps]
            ub = UnionBlock(elems)
            views1 = views1 + [ub]

        if views0 and views1:
            views1.insert(0, views0)
            return views1
        elif views0:
            return [views0]
        elif views1:
            return views1
        else:
            return []

    def decomposeBGP(self, tl):
        stars = self.getQueryStar(tl)

        selectedmolecules = {}
        varpreds = {}
        starpreds = {}
        conn = self.getStarsConnections(stars)
        splitedstars = {}

        for s in stars.copy():
            ltr = stars[s]
            preds = [utils.getUri(tr.predicate, self.prefixes)[1:-1] for tr in ltr if tr.predicate.constant]
            starpreds[s] = preds
            typemols = self.checkRDFTypeStatemnt(ltr)
            if len(typemols) > 0:
                selectedmolecules[s] = typemols
                for m in typemols:
                    properties = [p['predicate'] for p in self.config.metadata[m]['predicates']]
                    pinter = set(properties).intersection(preds)
                    if len(pinter) != len(preds):
                        print("Subquery: ", stars[s], "\nCannot be executed, because it contains properties that "
                                                      "does not exist in this federations of datasets.")
                        return []
                continue

            if len(preds) == 0:
                found = False
                for v in conn.values():
                    if s in v:
                        mols = [m for m in self.config.metadata]
                        found = True
                if not found:
                    varpreds[s] = ltr
                    continue
            else:
                mols = self.config.findbypreds(preds)

            if len(mols) > 0:
                if s in selectedmolecules:
                    selectedmolecules[s].extend(mols)
                else:
                    selectedmolecules[s] = mols
            else:
                splitstars = self.config.find_preds_per_mt(preds)
                if len(splitstars) == 0:
                    print("cannot find any matching molecules for:", tl)
                    return []
                else:
                    splitedstars[s] = [stars[s], splitstars, preds]
                    for m in list(splitstars.keys()):
                        selectedmolecules[str(s+'_'+m)] = [m]
                    #return self.decomposeSplitedStar(stars[s], splitstars, preds)

        if len(varpreds) > 0:
            mols = [m for m in self.config.metadata]
            for s in varpreds:
                selectedmolecules[s] = mols

        molConn = self.getMTsConnection(selectedmolecules)
        if len(splitedstars) > 0:
            for s in splitedstars:
                newstarpreds = {utils.getUri(tr.predicate, self.prefixes)[1:-1]: tr for tr in stars[s] if tr.predicate.constant}

                for m in splitedstars[s][1]:
                    stars[str(s + '_' + m)] = [newstarpreds[p] for p in splitedstars[s][1][m]]
                    starpreds[str(s + '_' + m)] = splitedstars[s][1][m]
                del stars[s]
                del starpreds[s]

        conn = self.getStarsConnections(stars)
        res = self.pruneMTs(conn, molConn, selectedmolecules, stars)
        # print(res)
        qpl0 = []
        qpl1 = []
        for s in res:
            if len(res[s]) == 1:
                if len(self.config.metadata[res[s][0]]['wrappers']) == 1:
                    endpoint = self.config.metadata[res[s][0]]['wrappers'][0]['url']
                    qpl0.append(Service("<" + endpoint + ">", list(set(stars[s]))))
                else:
                    sources = [w['url'] for w in self.config.metadata[res[s][0]]['wrappers']
                               if len(starpreds[s]) == len(list(set(starpreds[s]).intersection(set(w['predicates']))))]
                    # joins = [JoinBlock([Service("<" + url + ">", list(set(stars[s])))]) for url in sources]
                    # qpl1.append([UnionBlock([UnionBlock(joins)])])
                    if len(sources) == 1:
                        endpoint = sources[0]
                        qpl0.append(Service("<" + endpoint + ">", list(set(stars[s]))))
                    elif len(sources) > 1:
                        elems = [JoinBlock([Service("<" + ep + ">", list(set(stars[s])))]) for ep in sources]
                        ub = UnionBlock(elems)
                        qpl1.append(ub)
                    else:
                        # split and join
                        wpreds = {}

                        ptrs = {utils.getUri(tr.predicate, self.prefixes)[1:-1]: tr for tr in stars[s] if tr.predicate.constant}
                        for w in self.config.metadata[res[s][0]]['wrappers']:
                            wps = [p for p in w['predicates'] if p in starpreds[s]]
                            wpreds[w['url']] = wps

                        inall = []
                        difs = {}
                        for e in wpreds:
                            if len(inall) == 0:
                                inall = wpreds[e]
                            else:
                                inall = list(set(inall).intersection(wpreds[e]))

                            if e not in difs:
                                difs[e] = wpreds[e]
                            for d in difs:
                                if e == d:
                                    continue
                                dd = list(set(difs[d]).difference(wpreds[e]))
                                if len(dd) > 0:
                                    difs[d] = dd

                                dd = list(set(difs[e]).difference(wpreds[d]))
                                if len(dd) > 0:
                                    difs[e] = dd

                        oneone = {}
                        for e1 in wpreds:
                            for e2 in wpreds:
                                if e1 == e2 or e2 + '|-|' + e1 in oneone:
                                    continue
                                pp = set(wpreds[e1]).intersection(wpreds[e2])
                                pp = list(set(pp).difference(inall))
                                if len(pp) > 0:
                                    oneone[e1 + '|-|' + e2] = pp
                        onv = []
                        [onv.extend(d) for d in list(oneone.values())]
                        difv = []
                        [difv.extend(d) for d in list(difs.values())]
                        for o in onv:
                            if o in difv:
                                toremov = []
                                for d in difs:
                                    if o in difs[d]:
                                        difs[d].remove(o)
                                        difv.remove(o)
                                    if len(difs[d]) == 0:
                                        toremov.append(d)
                                for d in toremov:
                                    del difs[d]

                        ddd = onv + difv
                        rdftype = []
                        if len(set(inall + ddd)) == len(starpreds[s]):
                            if len(inall) > 0:
                                if len(inall) == 1 and inall[0] == 'http://www.w3.org/1999/02/22-rdf-syntax-ns#type':
                                    rdftype.extend(list(wpreds.keys()))
                                    pass
                                else:
                                    trps = [ptrs[p] for p in inall]
                                    elems = [JoinBlock([Service("<" + ep + ">", list(set(trps)))]) for ep in list(wpreds.keys())]
                                    ub = UnionBlock(elems)
                                    qpl1.append(ub)
                            if len(oneone) > 0:
                                for ee in oneone:
                                    e1, e2 = ee.split("|-|")
                                    pp = oneone[ee]
                                    if len(pp) == 1 and pp[0] == 'http://www.w3.org/1999/02/22-rdf-syntax-ns#type':
                                        rdftype.extend([e1, e2])
                                        pass
                                    else:
                                        trps = [ptrs[p] for p in pp]
                                        elems = [JoinBlock([Service("<" + e1 + ">", list(set(trps)))]),
                                                 JoinBlock([Service("<" + e2 + ">", list(set(trps)))])]
                                        ub = UnionBlock(elems)
                                        qpl1.append(ub)
                            if len(difs) > 0:
                                for d in difs:
                                    trps = [ptrs[p] for p in difs[d]]
                                    if d in rdftype:
                                        trps.append(ptrs['http://www.w3.org/1999/02/22-rdf-syntax-ns#type'])
                                    qpl0.append(Service("<" + d + ">", list(set(trps))))
                        else:
                            return []
            else:
                # preds = [utils.getUri(tr.predicate, self.prefixes)[1:-1] for tr in stars[s] if tr.predicate.constant]
                # mulres = self.decompose_multimolecule(res[s], stars[s], preds)
                # if isinstance(mulres, Service):
                #     results.append(mulres)
                # else:
                #     results.extend(mulres)

                md = self.metawrapperdecomposer(res[s], stars[s])
                if isinstance(md, Service):
                    qpl0.append(md)
                else:
                    for m in md:
                        if isinstance(m, Service):
                            qpl0.append(m)
                        else:
                            qpl1.append(m)
        if qpl0 and not self.joinlocally:
            joins = {}
            g = 0
            merged = []
            for i in range(len(qpl0)):
                if i+1 < len(qpl0):
                    for j in range(i+1, len(qpl0)):
                        s = qpl0[i]
                        k = qpl0[j]
                        if s.endpoint == k.endpoint:
                            if self.shareAtLeastOneVar(k.triples, s.triples):
                                if s.endpoint in joins:
                                    joins[s.endpoint].extend(s.triples + k.triples)
                                else:
                                    joins[s.endpoint] = s.triples + k.triples
                                merged.append(s)
                                merged.append(k)
                                joins[s.endpoint] = list(set(joins[s.endpoint]))

            [qpl0.remove(r) for r in set(merged)]
            for s in qpl0:
                if s.endpoint in joins:
                    if self.shareAtLeastOneVar(joins[s.endpoint], s.triples):
                        joins[s.endpoint].extend(s.triples)
                    else:

                        joins[s.endpoint + "|" + str(g)] = s.triples
                        g += 1
                else:
                    joins[s.endpoint] = s.triples

                joins[s.endpoint] = list(set(joins[s.endpoint]))

            qpl0 = []
            for e in joins:
                endp = e.split('|')[0]

                qpl0.append(Service('<' + endp + '>', joins[e]))

        if qpl0 and qpl1:
            qpl1.insert(0, qpl0)
            return qpl1
        elif qpl0 and not qpl1:
            return qpl0
        else:
            return qpl1

    def shareAtLeastOneVar(self, left, right):
        # joins can also be over constants, so also consider them
        leftsubj = [s.subject.name for s in left]
        leftobj = [s.theobject.name for s in left]
        rightsubj = [s.subject.name for s in right]
        rightobj = [s.theobject.name for s in right]

        leftvars = leftsubj + leftobj
        rightvars = rightsubj + rightobj
        inter = set(leftvars).intersection(set(rightvars))
        # print (inter)
        if len(inter) > 0:
            return True

        return False

    def metawrapperdecomposer(self, res, triplepatterns):
        sourceindex = dict()
        urlmoleculemap = dict()
        predtrips = dict()
        preds = []
        for tr in triplepatterns:
            if tr.predicate.constant:
                p = utils.getUri(tr.predicate, self.prefixes)[1:-1]
                predtrips[p] = tr
                preds.append(p)
        for x in res:
            wrappers = self.config.metadata[x]
            wrappers = [w for w in wrappers['wrappers']]
            if len(wrappers) > 1:
                for w in wrappers:
                    exitsingpreds = []
                    for p in preds:
                        if p in w['predicates']:
                            exitsingpreds.append(predtrips[p])
                    if len(exitsingpreds) == 0:
                        continue
                    urlmoleculemap[w['url']] = x
                    if w['url'] not in sourceindex:
                        sourceindex[w['url']] = exitsingpreds
                    else:
                        sourceindex[w['url']].extend(exitsingpreds)
                        sourceindex[w['url']] = list(set(sourceindex[w['url']]))
            else:
                exitsingpreds = []
                w = wrappers[0]
                for p in preds:
                    if p in w['predicates']:
                        exitsingpreds.append(predtrips[p])
                if len(exitsingpreds) == 0:
                    continue

                urlmoleculemap[w['url']] = x
                if w['url'] not in sourceindex:
                    sourceindex[w['url']] = exitsingpreds
                else:
                    sourceindex[w['url']].extend(exitsingpreds)
                    sourceindex[w['url']] = list(set(sourceindex[w['url']]))

        if len(sourceindex) == 1:
            return Service('<' + list(sourceindex.keys())[0] + '>', list(set(triplepatterns)))

        # for url in sourceindex:
        #     eps = sourceindex[url]
        #     if len(eps) == len(triplepatterns):
        #         return Service('<' + url + '>', list(set(triplepatterns))) #urlmoleculemap[url]

        intersects = None
        for url in sourceindex:
            if intersects is None:
                intersects = set(sourceindex[url])
                continue
            intersects = intersects.intersection(set(sourceindex[url]))
            if len(intersects) == 0:
                break

        joins = []
        servs = []
        if intersects and len(intersects) > 0:
            singlesource = {url: sourceindex[url] for url in sourceindex if len(sourceindex[url]) == len(triplepatterns)}
            if len(singlesource) > 0:
                if len(singlesource) == 1:
                    for url in singlesource:
                        servs.append(Service("<" + url + ">", list(set(sourceindex[url]))))
                else:
                    for url in singlesource:
                        joins.append(JoinBlock([Service("<" + url + ">", list(set(sourceindex[url])))]))
            else:
                [sourceindex[url].remove(e) for e in intersects for url in sourceindex]
                ignore = []
                for url in sourceindex:
                    if len(sourceindex[url]) == len(triplepatterns):
                        servs.append(Service("<" + url + ">", list(set(sourceindex[url]))))
                        ignore.append(url)
                    else:
                        joins.append(JoinBlock([Service("<" + url + ">", list(intersects))]))
                for url in sourceindex:
                    if len(sourceindex[url]) > 0 and url not in ignore:
                        servs.append(Service("<" + url + ">", list(set(sourceindex[url]))))
            # if len(servs) == len(sourceindex):
            #     joins = servs
            #     servs = []
            # elif len(servs) == 1:
            #     joins = []
        else:
            #TODO: check other decompositions to make a true union
            joins.extend([JoinBlock([Service("<" + url + ">", triplepatterns)]) for url in sourceindex])

        if len(joins) > 0:
            servs.append(UnionBlock(joins))
        return servs

    def getMTsConnection(self, selectedmolecules):
        mcons = {}
        smolecules = [m for s in selectedmolecules for m in selectedmolecules[s]]
        for s in selectedmolecules:
            mols = selectedmolecules[s]
            for m in mols:
                mcons[m] = [n for n in self.config.metadata[m]['linkedTo'] if n in smolecules]
        return mcons

    def pruneMTs(self, conn, molConn, selectedmolecules, stars):
        newselected = {}
        res = {}
        for s in selectedmolecules:
            if len(selectedmolecules[s]) == 1:
                newselected[s] = list(selectedmolecules[s])
                res[s] = list(selectedmolecules[s])
            else:
                newselected[s] = []
                res[s] = []

        for s in selectedmolecules:
            sc = conn[s]
            for sm in selectedmolecules[s]:
                smolink = molConn[sm]
                for c in sc:
                    cmols = selectedmolecules[c]
                    nms = [m for m in smolink if m in cmols]
                    if len(nms) > 0:
                        res[s].append(sm)
                        res[c].extend(nms)
        #check predicate level connections
        newfilteredonly = {}
        for s in res:
            sc = [c for c in conn if s in conn[c]]
            for c in sc:
                connectingtp = [utils.getUri(tp.predicate, self.prefixes)[1:-1]
                         for tp in stars[c] if tp.theobject.name == s]
                connectingtp = list(set(connectingtp))
                sm = selectedmolecules[s]
                for m in sm:
                    srange = [p for r in self.config.metadata[m]['predicates'] for p in r['range'] if
                              r['predicate'] in connectingtp]
                    filteredmols = [r for r in res[s] if r in srange]
                    if len(filteredmols) > 0:
                        if s in newfilteredonly:
                            newfilteredonly[s].extend(filteredmols)
                        else:
                            res[s] = filteredmols

        for s in newfilteredonly:
            res[s] = list(set(newfilteredonly[s]))

        for s in res:
            if len(res[s]) == 0:
                res[s] = selectedmolecules[s]
            res[s] = list(set(res[s]))
        return res

    def checkRDFTypeStatemnt(self, ltr):
        types = self.getRDFTypeStatement(ltr)
        typemols = []
        for t in types:
            tt = utils.getUri(t.theobject, self.prefixes)[1:-1]
            if tt in self.config.metadata:
                typemols.append(tt)

        return typemols

    def getStarsConnections(self, stars):
        """
        extracts links between star-shaped sub-queries
        :param stars: map of star-shaped sub-queries with its root (subject) {subject: [triples in BGP]}
        :return: map of star-shaped sub-query root name (subject) with its connected sub-queries via its object node.
         {subj1: [subjn]} where one of subj1's triple pattern's object node is connected to subject node of subjn
        """
        conn = dict()
        for s in stars.copy():
            ltr = stars[s]
            conn[s] = []
            for c in stars:
                if c == s:
                    continue
                for t in ltr:
                    if t.theobject.name == c:
                        if c not in conn[s]:
                            conn[s].append(c)
                        break

        return conn
    '''
    ===================================================
    ========= STAR-SHAPED DECOMPOSITIONS ==============
    ===================================================
    '''
    def getQueryStar(self, tl):
        """
        extracts star-shaped subqueries from a list of triple patterns in a BGP
        :param tl: list of triple patterns in a BGP
        :return: map of star-shaped sub-queries with its root (subject) {subject: [triples in BGP]}
        """
        stars = dict()
        for t in tl:
            if t.subject.name in stars:
                stars[t.subject.name].append(t)
            else:
                stars[t.subject.name] = [t]
        return stars

    def getRDFTypeStatement(self, ltr):
        types = []
        for t in ltr:
            if t.predicate.constant \
                    and (t.predicate.name == "a"
                         or t.predicate.name == "rdf:type"
                         or t.predicate.name == "<http://www.w3.org/1999/02/22-rdf-syntax-ns#type>") \
                    and t.theobject.constant:
                types.append(t)

        return types

    '''
    ===================================================
    ========= FILTERS =================================
    ===================================================
    '''
    def includeFilter(self, jb_triples, fl):
        fl1 = []
        for jb in jb_triples:

            if isinstance(jb, list):
                for f in fl:
                    fl2 = self.includeFilterAux(f, jb)
                    fl1 = fl1 + fl2
            elif (isinstance(jb, UnionBlock)):
                for f in fl:
                    fl2 = self.includeFilterUnionBlock(jb, f)
                    fl1 = fl1 + fl2
            elif (isinstance(jb, Service)):
                for f in fl:
                    fl2 = self.includeFilterAuxSK(f, jb.triples, jb)
                    fl1 = fl1 + fl2
        return fl1

    def includeFilterAux(self, f, sl):
        fl1 = []
        for s in sl:
            # print(type(s))
            vars_s = set()
            for t in s.triples:
                vars_s.update(set(utils.getVars(t)))
            vars_f = f.getVars()
            if set(vars_s) & set(vars_f) == set(vars_f):
                s.include_filter(f)
                fl1 = fl1 + [f]
            elif type(f) is Values and set(vars_s) & set(vars_f) != set():
                fl2 = f.instantiate(set(vars_s) & set(vars_f))
                s.include_filter(fl2) # the new decomposed clause need to be reconsidered
                fl1 = fl1 + [fl2]
        return fl1

    def includeFilterUnionBlock(self, jb, f):
        fl1 = []
        for jbJ in jb.triples:
            for jbUS in jbJ.triples:
                if isinstance(jbUS, Service):
                    vars_s = set(jbUS.getVars())
                    vars_f = f.getVars()
                    if type(f) is Values and set(vars_s) & set(vars_f) != set():
                        fl2 = f.instantiate(set(vars_s) & set(vars_f))
                        jbUS.include_filter(fl2)
                        fl1 = fl1 + [fl2]
                    elif set(vars_s) & set(vars_f) == set(vars_f):
                        jbUS.include_filter(f)
                        fl1 = fl1 + [f]
        return fl1

    def includeFilterAuxSK(self, f, sl, sr):
        """
        updated: includeFilterAuxS(f, sl, sr) below to include filters that all vars in filter exists in any of the triple
        patterns of a BGP. the previous impl includes them only if all vars are in a single triple pattern
        :param f:
        :param sl:
        :param sr:
        :return:
        """
        fl1 = []
        serviceFilter = False
        fvars = dict()
        vars_f = f.getVars()

        for v in vars_f:
            fvars[v] = False
        bgpvars = set()

        if not isinstance(sl, list):
            sl = [sl]

        for s in sl:
            bgpvars.update(set(utils.getVars(s)))
            vars_s = set()
            if (isinstance(s, Triple)):
                vars_s.update(set(utils.getVars(s)))
            else:
                for t in s.triples:
                    vars_s.update(set(utils.getVars(t)))

            if set(vars_s) & set(vars_f) == set(vars_f): # Bind supposed to fall to this category
                serviceFilter = True

            if type(f) is Values and set(vars_s) & set(vars_f) != set():
                fl2 = f.instantiate(set(vars_s) & set(vars_f))
                serviceFilter = True

        for v in bgpvars:
            if v in fvars:
                fvars[v] = True

        # if type(f) is Bind: # why is this here? Investigate later
        #     fvars[f.alias] = True

        if serviceFilter:
            if type(f) is Values:
                sr.include_filter(fl2)
                fl1 = fl1 + [fl2]
            else:
                sr.include_filter(f)
                fl1 = fl1 + [f]
        else:
            fs = [v for v in fvars if not fvars[v]]
            if len(fs) == 0:
                sr.include_filter(f)
                fl1 = fl1 + [f]
        return fl1

    def updateFilters(self, node, filters):
        return UnionBlock(node.triples, filters)

    '''
    ===================================================
    ========= MAKE PLAN =================================
    ===================================================
    '''
    def makePlanQuery(self, q):
        x = self.makePlanUnionBlock(q.body)
        return x

    def makePlanUnionBlock(self, ub):
        r = []
        for jb in ub.triples:
            r.append(self.makePlanJoinBlock(jb))
        return UnionBlock(r, ub.filters)

    def makePlanJoinBlock(self, jb):
        sl = []
        ol = []

        for bgp in jb.triples:
            if type(bgp) == list:
                sl.extend(bgp)
            elif isinstance(bgp, Optional):

                for f in jb.filters:
                    vars_f = f.getVars()
                    if set(bgp.getVars()) & set(vars_f) == set(vars_f):
                        for t in bgp.bgg.triples:
                            if set(t.getVars()) & set(vars_f) == set(vars_f):
                                t.filters.extend(jb.filters)

                ol.append(Optional(self.makePlanUnionBlock(bgp.bgg)))
            elif isinstance(bgp, UnionBlock):

                for f in jb.filters:
                    vars_f = f.getVars()
                    if set(bgp.getVars()) & set(vars_f) == set(vars_f):
                        for t in bgp.triples:
                            if set(t.getVars()) & set(vars_f) == set(vars_f):
                                t.filters.extend(jb.filters)

                sl.append(self.makePlanUnionBlock(bgp))
            elif isinstance(bgp, JoinBlock):

                for f in jb.filters:
                    vars_f = f.getVars()
                    if set(bgp.getVars()) & set(vars_f) == set(vars_f):
                        bgp.filters.extend(jb.filters)

                sl.append(self.makePlanJoinBlock(bgp))
            elif isinstance(bgp, Service):

                for f in jb.filters:
                    vars_f = f.getVars()
                    if set(bgp.getVars()) & set(vars_f) == set(vars_f):
                        bgp.filters.extend(jb.filters)

                sl.append(bgp)

        pl = self.makePlanAux(sl, jb.filters)
        if ol:
            pl = [pl]
            pl.extend(ol)

        return JoinBlock(pl, filters=jb.filters)

    def makePlanAux(self, ls, filters=None):
        if filters is None:
            filters = []
        return self.makeBushyTree(ls, filters)

    def makeBushyTree(self, ls, filters=None):
        if filters is None:
            filters = []
        return Tree.makeBushyTree(ls, filters)

    def makeNaiveTree(self, ls):
        return Tree.makeNaiveTree(ls)

    def makeLeftLinearTree(self, ls):
        return Tree.makeLLTree(ls)


class DecompositionType(Enum):

    STAR = partial(Decomposer.decomposeBGP)
    EG = partial(Decomposer.decompose_exclusive_groups)
    TRIPLE = partial(Decomposer.decompose_triple_wise)

    def decompose(self, decomposer, tl):
        return self.value(decomposer, tl)



"""
Created on Jul 10, 2011

Implements the Xgoptional operator.
The intermediate results are represented as queues.

@author: Maribel Acosta Deibe
"""

from multiprocessing import Queue
from time import time
from DeTrusty.Operators.Optional import Optional
from .OperatorStructures import Record, RJTTail


class Xgoptional(Optional):

    def __init__(self, vars_left, vars_right):
        self.left_table  = dict()
        self.right_table = dict()
        self.qresults    = Queue()
        self.bag         = []
        self.vars_left   = set(vars_left)
        self.vars_right  = set(vars_right)
        self.vars        = list(self.vars_left & self.vars_right)

    def instantiate(self, d):
        newvars_left = self.vars_left - set(d.keys())
        newvars_right = self.vars_right - set(d.keys())
        return Xgoptional(newvars_left, newvars_right)

    def instantiateFilter(self, instantiated_vars, filter_str):

        newvars_left = self.vars_left - set(instantiated_vars)
        newvars_right = self.vars_right - set(instantiated_vars)
        return Xgoptional(newvars_left, newvars_right)

    def execute(self, left, right, out, processqueue=Queue()):
        # Executes the Xgoptional.
        self.left     = left
        self.right    = right
        self.qresults = out

        # Initialize tuples.
        tuple1 = None
        tuple2 = None

        # Get the tuples from the queues.
        while (not(tuple1 == "EOF") or not(tuple2 == "EOF")):
            # Try to get and process tuple from left queue.

            if (not(tuple1 == "EOF")):

                try:
                    tuple1 = self.left.get(False)
                    if not(tuple1 == "EOF"):
                        self.bag.append(tuple1)
                    self.stage1(tuple1, self.left_table, self.right_table, self.vars_right)
                    #print  "bag after stage 1: ", self.bag
                except Exception  as e:
                    # This catch:
                 #   print "Exception ", e.message
                    # Empty: in tuple2 = self.right.get(False), when the queue is empty.
                    # TypeError: in resource = resource + tuple[var], when the tuple is "EOF".
                    pass

            # Try to get and process tuple from right queue.
            if (not(tuple2 == "EOF")):

                try:
                    tuple2 = self.right.get(False)
                    if not(tuple2 == "EOF"):
                        self.stage1(tuple2, self.right_table, self.left_table, self.vars_left)
                except Exception:
                    # This catch:
                    # Empty: in tuple2 = self.right.get(False), when the queue is empty.
                    # TypeError: in resource = resource + tuple[var], when the tuple is "EOF".
                    pass

        #print " Perform the last probes."
        self.stage3()


    def stage1(self, tuple, tuple_rjttable, other_rjttable, vars):
        # Stage 1: While one of the sources is sending data.
        #print "stage 1"
        # Get the resource associated to the tuples.
        resource = ''
        for var in self.vars:
            resource = resource + tuple[var]
        #print "probe"
        # Probe the tuple against its RJT table.
        probeTS = self.probe(tuple, resource, tuple_rjttable, vars)
        #print "creating record"
        # Create the records.
        record = Record(tuple, probeTS, time())
        #print "Stage 1"
        # Insert the record in the other RJT table.
        # TODO: use RJTTail. Check ProbeTS
        if resource in other_rjttable:
            other_rjttable.get(resource).updateRecords(record)
            other_rjttable.get(resource).setRJTProbeTS(probeTS)
            #other_rjttable.get(resource).append(record)
        else:
            tail = RJTTail(record, float("inf"))
            other_rjttable[resource] = tail
            #other_rjttable[resource] = [record]

    #def stage2(self):
        # Stage 2: When both sources become blocked.
        #pass

    def stage2(self, signum, frame):
        # print " Stage 2: When both sources become blocked."
        self.sourcesBlocked = True

        # Get common resources.
        resources1 = set(self.left_table.keys()) & set(self.fileDescriptor_right.keys())
        resources2 = set(self.right_table.keys()) & set(self.fileDescriptor_left.keys())

        # Iterate while there are common resources and both sources are blocked.
        while ((resources1 or resources2) and self.sourcesBlocked):

            if (resources1):
                resource = resources1.pop()
                rjts1 = self.left_table[resource].records
                for rjt1 in rjts1:
                    probed = self.probeFile(rjt1, self.fileDescriptor_right, resource, 2)
                    if (probed):
                        rjt1.probeTS = time()

            elif (resources2):
                resource = resources2.pop()
                rjts1 = self.right_table[resource].records
                for rjt1 in rjts1:
                    probed = self.probeFile(rjt1, self.fileDescriptor_left, resource, 2)
                    if (probed):
                        rjt1.probeTS = time()

        # End of second stage.
        self.lastSecondStageTS = time()
        self.secondStagesTS.append(self.lastSecondStageTS)

    def stage3(self):
        # Stage 3: When both sources sent all the data.
        #print "stage 3"
        # This is the optional: Produce tuples that haven't matched already.
        #print "Length of data in xgoptional(): ", len(self.bag)
        for tuple in self.bag:
            #print 'stage 3 loop in bag'
            res_right = {}
            for var in self.vars_right:
                res_right.update({var:''})
            res = res_right
            res.update(tuple)
            self.qresults.put(res)

        # Put EOF in queue and exit.
        self.qresults.put("EOF")
        return

    def probe(self, tuple, resource, rjttable, vars):
        probeTS = time()
        # If the resource is in table, produce results.
        try:
            if resource in rjttable:
                rjttable.get(resource).setRJTProbeTS(probeTS)
                list_records = rjttable[resource].records
                # Delete tuple from bag.
                try:
                    self.bag.remove(tuple)
                except ValueError:
                    pass

                for record in list_records:
                    #print "record: ", type(record.tuple), record.tuple
                    if isinstance(record.tuple, dict):
                        res = record.tuple.copy()
                        res.update(tuple)
                        self.qresults.put(res)

                        # Delete tuple from bag.
                        try:
                            self.bag.remove(record.tuple)
                        except ValueError:
                            pass
        except Exception as e:
            pass
        return probeTS



"""
Created on 10/12/2021.

Implements the Xaggregate operator.
The intermediate results are represented in a queue.

@author: Avellino
"""

from DeTrusty.Sparql.Parser.services import getPrefs, Expression
from multiprocessing import Queue
import operator
import datetime

unary_operators = {
    '!': operator.not_,
    '+': '',
    '-': operator.neg
}

logical_connectives = {
    '||': operator.or_,
    '&&': operator.and_
}

arithmetic_operators = {
    '*': operator.mul,
    '/': operator.truediv,
    '+': operator.add,
    '-': operator.sub
}

test_operators = {
    '=': operator.eq,
    '!=': operator.ne,
    '<': operator.lt,
    '>': operator.gt,
    '<=': operator.le,
    '>=': operator.ge
}

data_types = {
    'integer': (int, 'numerical'),
    'decimal': (float, 'numerical'),
    'float': (float, 'numerical'),
    'double': (float, 'numerical'),
    'string': (str, str),
    'boolean': (bool, bool),
    'dateTime': (datetime, datetime),
    'nonPositiveInteger': (int, 'numerical'),
    'negativeInteger': (int, 'numerical'),
    'long': (int, 'numerical'),
    'int': (int, 'numerical'),
    'short': (int, 'numerical'),
    'byte': (bytes, bytes),
    'nonNegativeInteger': (int, 'numerical'),
    'unsignedLong': (int, 'numerical'),
    'unsignedInt': (int, 'numerical'),
    'unsignedShort': (int, 'numerical'),
    'unsignedByte': (bytes, bytes)  # TODO: this is not correct
}

numerical = (int, float)

class Xaggregate(object):

    def __init__(self, args, over_all_triples, group_args, prefs):
        self.qresults = Queue()
        self.ban_list = list()
        self.one_res = dict()
        self.args = args
        self.prefsDict = getPrefs(prefs)
        
        for arg in args:
            self.ban_list.append(arg.getName())
            if over_all_triples:
                self.one_res.update({arg.getName(): arg.getDefVal()})

        for arg in group_args: # TODO: adjust after GROUP BY parser accepts Expression
            if arg.alias:
                self.ban_list.append(arg.alias[1:])
            else:
                self.ban_list.append(arg.name[1:])

    def execute(self, left, dummy, out, processqueue=Queue()):
        self.left = left
        self.qresults = out
        tuple = self.left.get(True)
        
        while tuple != "EOF":
            for arg in self.args:
                if type(arg.exp) is Expression:
                    self.exec_simplifyExp(arg.exp, tuple) # TODO: update tuple with return value
                else:
                    self.exec_simplifiedExp(arg.exp, arg.distinct, arg.op, arg.getName(), arg.sep, tuple)
            if not self.one_res: 
                self.qresults.put(tuple)
            tuple = self.left.get(True)

        if self.one_res:
            for arg in self.args:

                if arg.op.upper() == 'COUNT':
                    self.one_res.update({arg.getName(): str(len(self.one_res[arg.getName()]))})

                elif arg.op.upper() == 'GROUP_CONCAT':
                    if self.one_res[arg.getName()]:
                        tmp = str(self.one_res[arg.getName()])
                        if arg.sep:
                            tmp1 = tmp.replace("'", "")
                            ret = tmp1.replace(", ", arg.sep[1:len(arg.sep)-1])
                        else:
                            ret = tmp.replace("'", "")
                        self.one_res.update({arg.getName(): ret[1:len(ret)-1]})
                    else:
                        self.one_res.update({arg.getName(): ''})

                elif arg.op.upper() == 'SAMPLE':
                    if self.one_res[arg.getName()]:
                        self.one_res.update({arg.getName(): str(self.one_res[arg.getName()])})
                    else:
                        self.one_res.update({arg.getName(): ''})

                elif arg.op.upper() == 'SUM':
                    if self.one_res[arg.getName()]:
                        ret = 0
                        str_in_ls = False
                        for val in self.one_res[arg.getName()]:
                            if val.isdigit():
                                ret += int(val)
                            elif val.replace('.','', 1).isdigit():
                                ret += float(val)
                            else:
                                str_in_ls = True
                        if str_in_ls:
                            self.one_res.update({arg.getName(): ''})
                        else:
                            self.one_res.update({arg.getName(): str(ret)})
                    else:
                        self.one_res.update({arg.getName(): ''})

                elif arg.op.upper() == 'MIN':
                    if self.one_res[arg.getName()]:
                        self.one_res[arg.getName()].sort()
                        self.one_res.update({arg.getName(): str(self.one_res[arg.getName()][0])})
                    else:
                        self.one_res.update({arg.getName(): ''})

                elif arg.op.upper() == 'MAX':
                    if self.one_res[arg.getName()]:
                        self.one_res[arg.getName()].sort(reverse=True)
                        self.one_res.update({arg.getName(): str(self.one_res[arg.getName()][0])})
                    else:
                        self.one_res.update({arg.getName(): ''})

                elif arg.op.upper() == 'AVG':
                    if self.one_res[arg.getName()]:
                        ret = 0
                        str_in_ls = False
                        for val in self.one_res[arg.getName()]:
                            if val.isdigit():
                                ret += int(val)
                            elif val.replace('.', '', 1).isdigit():
                                ret += float(val)
                            else:
                                str_in_ls = True
                        if str_in_ls:
                            self.one_res.update({arg.getName(): ''})
                        else:
                            self.one_res.update({arg.getName(): str(ret/len(self.one_res[arg.getName()]))})
                    else:
                        self.one_res.update({arg.getName(): ''})

            self.qresults.put(self.one_res)
        self.qresults.put("EOF")
        return

    def exec_simplifyExp(self, arg, tup):
        if type(arg) is Expression:
            if arg.op in arithmetic_operators:
                return self.evaluateAritmethic(arg, tup)
            elif arg.op in unary_operators:
                return self.evaluateUnaryOperator(arg, tup)
            elif arg.op in test_operators:
                return self.evaluateTest(arg, tup)
            elif arg.op in logical_connectives:
                return self.evaluateLogicalConnective(arg)
        else:
            if arg.constant:
                if (arg.name[0] == '<' and arg.name[len(arg.name)-1] == '>') or ((arg.name[0] == "'" and arg.name[len(arg.name)-1] == "'") or (arg.name[0] == '"' and arg.name[len(arg.name)-1] == '"')):
                    return arg.name[1:len(arg.name)-1]
                elif ":" in arg.name:
                    pos = arg.name.find(":")
                    c = arg.name[0:pos].strip()
                    v = arg.name[(pos+1):len(arg.name)].strip()
                    if c in self.prefsDict:
                        return self.prefsDict[c][1:len(self.prefsDict[c])-1] + v
                    else: # TODO: throw exception
                        pass
                else:
                    return arg.name
            else:
                return tup[arg.name[1:]]

    def evaluateAritmethic(self, arg, tup):
        term_left = str(self.exec_simplifyExp(arg.left, tup))
        term_right = str(self.exec_simplifyExp(arg.right, tup))
        if term_left.isdigit() and term_right.isdigit():
            term_left = int(term_left)
            term_right = int(term_right)
            return (arithmetic_operators[operator](term_left, term_right))
        elif term_right.replace('.','',1).isdigit() and term_left.isdigit():
            term_right = float(term_right)
            term_left = int(term_left)
            return (arithmetic_operators[operator](term_left, term_right))
        elif term_left.replace('.','',1).isdigit() and term_right.isdigit():
            term_left = float(term_left)
            term_right = int(term_right)
            return (arithmetic_operators[operator](term_left, term_right))
        elif term_left.replace('.','',1).isdigit() and term_right.replace('.','',1).isdigit():
            term_left = float(term_left)
            term_right = float(term_right)
            return (arithmetic_operators[operator](term_left, term_right))
        else:
            raise SPARQLTypeError

    def evaluateUnaryOperator(self, arg, tup):
        term_left = exec_simplifyExp(arg.left, tup)
        return unary_operators[arg.op](term_left)

    def evaluateTest(self, arg, tup):
        left = exec_simplifyExp(arg.left)
        right = exec_simplifyExp(arg.right)
        (term_left, type_left) = left, type(left)
        (term_right, type_right) = right, type(right)
        term_left = str(term_left)
        term_right = str(term_right)
        if term_left.isnumeric() and term_right.isnumeric():
            term_left = int(term_left)
            term_right = int(term_right)
            return (test_operators[operator](term_left, term_right), bool)
        elif term_left.replace('.','',1).isdigit() and term_right.isnumeric():
            term_left = float(term_left)
            term_right = int(term_right)
            return (test_operators[operator](term_left, term_right), bool)
        elif term_left.isnumeric() and term_right.replace('.','',1).isdigit():
            term_left = int(term_left)
            term_right = float(term_right)
            return (test_operators[operator](term_left, term_right), bool)
        elif term_left.replace('.','',1).isdigit() and term_right.replace('.','',1).isdigit():
            term_left = float(term_left)
            term_right = float(term_right)
            return (test_operators[operator](term_left, term_right), bool)
        else:
            print("SPARQLTypeError - in Xbind")
            raise SPARQLTypeError

    def exec_simplifiedExp(self, arg, dist, op, name, sep, tup):
        if arg.constant:
            if (arg.name[0] == '<' and arg.name[len(arg.name)-1] == '>') or ((arg.name[0] == "'" and arg.name[len(arg.name)-1] == "'") or (arg.name[0] == '"' and arg.name[len(arg.name)-1] == '"')):
                tup.update({name : arg.name[1:len(arg.name)-1]})
                if self.one_res:
                    if op.upper() == 'SAMPLE':
                        self.one_res[name] = arg.name[1:len(arg.name)-1]
                    elif dist:
                        if arg.name[1:len(arg.name)-1] not in self.one_res[name]:
                            self.one_res[name].append(arg.name[1:len(arg.name)-1])
                    else:
                        self.one_res[name].append(arg.name[1:len(arg.name)-1])
            elif ":" in arg.name:
                pos = arg.name.find(":")
                c = arg.name[0:pos].strip()
                v = arg.name[(pos+1):len(arg.name)].strip()
                if c in self.prefsDict:
                    tup.update({name : self.prefsDict[c][1:len(self.prefsDict[c])-1] + v})
                    if self.one_res:
                        if op.upper() == 'SAMPLE':
                            self.one_res[name] = self.prefsDict[c][1:len(self.prefsDict[c])-1] + v                        
                        elif dist:
                            if (self.prefsDict[c][1:len(self.prefsDict[c])-1] + v) not in self.one_res[name]:
                                self.one_res[name].append(self.prefsDict[c][1:len(self.prefsDict[c])-1] + v)
                        else:
                            self.one_res[name].append(self.prefsDict[c][1:len(self.prefsDict[c])-1] + v)
                else: # TODO: throw exception
                    pass
            else:
                tup.update({name : arg.name})
                if self.one_res:
                    if op.upper() == 'SAMPLE':
                        self.one_res[name] = arg.name
                    elif dist:
                        if arg.name not in self.one_res[name]:
                            self.one_res[name].append(arg.name)
                    else:
                        self.one_res[name].append(arg.name)
        else:
            if self.one_res:
                if arg.name == '?ALL':
                    if dist:
                        if tup not in self.one_res[name]:
                            self.one_res[name].append(tup)
                    else:
                        self.one_res[name].append(tup)
                elif tup[arg.name[1:]] != '':
                    if op.upper() == 'SAMPLE':
                        if not self.one_res[name]:
                            self.one_res[name] = tup[arg.name[1:]]
                    else:
                        if dist:
                            if tup[arg.name[1:]] not in self.one_res[name]:
                                self.one_res[name].append(tup[arg.name[1:]])
                        else:
                            self.one_res[name].append(tup[arg.name[1:]])
            else:
                if op.upper() == 'COUNT':
                    if arg.name == '?ALL': # TODO: how to handle distinct?
                        for arg in tup:
                            if arg not in self.ban_list:
                                if type(tup[arg]) == list:
                                    tup.update({name: str(len(tup[arg]))})
                                    break
                                else:
                                    tup.update({name: '1'})
                                    break

                    else:
                        if dist:
                            if type(tup[arg.name[1:]]) is list:
                                rec = list()
                                count = 0
                                for val in tup[arg.name[1:]]:
                                    if val not in rec and val != '':
                                        rec.append(val)
                                        count += 1
                                tup.update({name: str(count)})
                            else:
                                if tup[arg.name[1:]] == '':
                                    tup.update({name: '0'})
                                else:
                                    tup.update({name: '1'})
                                    
                        else:
                            if type(tup[arg.name[1:]]) is list:
                                count = 0
                                for val in tup[arg.name[1:]]:
                                    if val != '':
                                        count += 1
                                tup.update({name: str(count)})
                            else:
                                if tup[arg.name[1:]] == '':
                                    tup.update({name: '0'})
                                else:
                                    tup.update({name: '1'})

                elif op.upper() == 'GROUP_CONCAT':
                    if dist:
                        if type(tup[arg.name[1:]]) is list:
                            tmp = set(tup[arg.name[1:]])
                            if '' in tmp:
                                tmp.remove('')
                            if len(tmp) > 0:
                                tmp1 = str(tmp)
                                if sep:
                                    tmp2 = tmp1.replace("'", "")
                                    ret = tmp2.replace(", ", sep[1:len(sep)-1])
                                else:
                                    ret = tmp1.replace("'", "")
                                tup.update({name: ret[1:len(ret)-1]})
                            else:
                                tup.update({name: ''})
                        else:
                            tup.update({name: tup[arg.name[1:]]})
                    else:
                        if type(tup[arg.name[1:]]) is list:
                            tmp = tup[arg.name[1:]].copy()
                            while '' in tmp:
                                tmp.remove('')
                            if len(tmp) > 0:
                                tmp1 = str(tmp)
                                if sep:
                                    tmp2 = tmp1.replace("'", "")
                                    ret = tmp2.replace(", ", sep[1:len(sep)-1])
                                else:
                                    ret = tmp1.replace("'", "")
                                tup.update({name: ret[1:len(ret)-1]})
                            else:
                                tup.update({name: ''})
                        else:
                            tup.update({name: tup[arg.name[1:]]})

                elif op.upper() == 'SAMPLE':
                    if type(tup[arg.name[1:]]) is list:
                        tmp_ls = tup[arg.name[1:]]
                        found_non_empty = False
                        for val in tmp_ls:
                            if val != '':
                                tup.update({name: val})
                                found_non_empty = True
                                break
                        if not found_non_empty:
                            tup.update({name: ''})
                    else:
                        tup.update({name: tup[arg.name[1:]]})

                elif op.upper() == 'SUM':
                    if type(tup[arg.name[1:]]) is list:
                        str_in_ls = False
                        ret = 0
                        if dist:
                            tmp = list()
                            for val in tup[arg.name[1:]]:
                                if val != '' and val not in tmp:
                                    tmp.append(val)
                                    if val.isdigit():
                                        ret += int(val)
                                    elif val.replace('.','', 1).isdigit():
                                        ret += float(val)
                                    else:
                                        str_in_ls = True
                        else:
                            for val in tup[arg.name[1:]]:
                                if val != '':
                                    if val.isdigit():
                                        ret += int(val)
                                    elif val.replace('.','', 1).isdigit():
                                        ret += float(val)
                                    else:
                                        str_in_ls = True
                        if str_in_ls:
                            tup.update({name: ''})
                        else:
                            tup.update({name: str(ret)})
                    else:
                        is_num = True
                        if not tup[arg.name[1:]].replace('.', '', 1).isdigit():
                            is_num = False
                        if is_num:
                            tup.update({name: tup[arg.name[1:]]})
                        else:
                            tup.update({name: ''})
                                
                elif op.upper() == 'MIN':
                    if type(tup[arg.name[1:]]) is list:
                        tmp = None
                        if dist:
                            tmp = set(tup[arg.name[1:]])
                            if '' in tmp:
                                tmp.remove('')
                        else:
                            tmp = tup[arg.name[1:]].copy()
                            while '' in tmp:
                                tmp.remove('')
                        tmp.sort()
                        if tmp:
                            tup.update({name: str(tmp[0])})
                        else:
                            tup.update({name: ''})
                    else:
                        tup.update({name: tup[arg.name[1:]]})

                elif op.upper() == 'MAX':
                    if type(tup[arg.name[1:]]) is list:
                        tmp = None
                        if dist:
                            tmp = set(tup[arg.name[1:]])
                            if '' in tmp:
                                tmp.remove('')
                        else:
                            tmp = tup[arg.name[1:]].copy()
                            while '' in tmp:
                                tmp.remove('')
                        tmp.sort(reverse=True)
                        if tmp:
                            tup.update({name: str(tmp[0])})
                        else:
                            tup.update({name: ''})
                    else:
                        tup.update({name: tup[arg.name[1:]]})

                elif op.upper() == 'AVG':
                    if type(tup[arg.name[1:]]) is list:
                        str_in_ls = False
                        ret = 0
                        val_count = 0
                        if dist:
                            for val in tup[arg.name[1:]]:
                                if val != '' and val not in tmp:
                                    tmp.append(val)
                                    if val.isdigit():
                                        ret += int(val)
                                    elif val.replace('.','', 1).isdigit():
                                        ret += float(val)
                                    else:
                                        str_in_ls = True
                                    val_count += 1
                        else:
                            for val in tup[arg.name[1:]]:
                                if val != '':
                                    if val.isdigit():
                                        ret += int(val)
                                    elif val.replace('.','', 1).isdigit():
                                        ret += float(val)
                                    else:
                                        str_in_ls = True
                                    val_count += 1
                        if str_in_ls:
                            tup.update({name: ''})
                        else:
                            ret = ret / val_count
                            tup.update({name: str(ret)})
                    else:
                        is_num = True
                        if not tup[arg.name[1:]].replace('.', '', 1).isdigit():
                            is_num = False
                        if is_num:
                            tup.update({name: tup[arg.name[1:]]})
                        else:
                            tup.update({name: ''})



from __future__ import annotations

__author__ = 'Philipp D. Rohde and Kemele M. Endris'

import json
import logging
import multiprocessing
import sys
from queue import Queue
from time import time
from typing import Optional

from rdflib import Graph

from DeTrusty.Logger import get_logger
from DeTrusty.Molecule.MTManager import JSONConfig, MTCreationConfig
from DeTrusty.Wrapper.RDFWrapper import contact_source

logger = get_logger('rdfmts', '.rdfmts.log', file_and_console=True)

CONFIG = MTCreationConfig()
DEFAULT_OUTPUT_PATH = '/DeTrusty/Config/rdfmts.json'
TYPING_WIKIDATA = '<http://www.wikidata.org/prop/direct/P31>'

metas = [
    'http://www.w3.org/ns/sparql-service-description',
    'http://www.openlinksw.com/schemas/virtrdf#',
    'http://www.w3.org/2000/01/rdf-schema#',
    'http://www.w3.org/1999/02/22-rdf-syntax-ns#',
    'http://purl.org/dc/terms/Dataset',
    'http://www.w3.org/2002/07/owl#',
    'http://www.w3.org/2001/XMLSchema#',
    'nodeID://'
]


class Endpoint:
    """Simple representation of an endpoint. URL is mandatory but an endpoint might have optional parameters."""
    def __init__(self, url: str, params: dict = None):
        self.url = url
        self.params = params if params is not None else {}

    @property
    def types(self):
        return self.params.get('types', [])

    def get_params(self):
        if self.params is None or len(self.params.keys()) == 0:
            return ''
        params_public = self.params.copy()
        params_public.pop('token', None)
        params_public.pop('valid_until', None)
        params_public.pop('mappings', None)
        params_public.pop('types', None)
        return params_public


def create_rdfmts(endpoints: list | dict,
                  output: Optional[str] = DEFAULT_OUTPUT_PATH,
                  interlinking: bool = False) -> Optional[JSONConfig]:
    logger_wrapper = get_logger('DeTrusty.Wrapper.RDFWrapper')
    logger_wrapper.setLevel(logging.WARNING)  # temporarily disable logging of contacting the source

    if output is not None:
        try:
            with open(output, 'a') as f:
                if not f.writable():
                    raise PermissionError
        except FileNotFoundError as e:
            logger.critical('No such file or directory: ' + output + '\tMake sure the directory exists!')
            raise e
        except PermissionError as e:
            logger.critical('You may not have permissions to open or write to the file: ' + output +
                            '\tPlease, check your permissions!')
            raise e

    dsrdfmts = {}
    sparqlendps = {}
    eoffs = {}
    epros = []
    start = time()

    CONFIG.setEndpoints(endpoints)

    endpoints = [Endpoint(key, value) for key, value in CONFIG.endpoints.items()]
    endpoints = _accessible_endpoints(endpoints)
    if len(endpoints) == 0:
        logger.critical('None of the endpoints can be accessed. Please check if you write URLs properly!')
        sys.exit(1)
    for e in endpoints:
        tq = multiprocessing.Queue()
        eoffs[e.url] = tq
        p1 = multiprocessing.Process(target=_collect_rdfmts_from_source, args=(e, tq,))
        epros.append(p1)
        p1.start()

    while len(eoffs) > 0:
        for url in eoffs:
            q = eoffs[url]
            rdfmts = q.get()
            sparqlendps[url] = rdfmts
            del eoffs[url]
            break
    for p in epros:
        if p.is_alive():
            p.terminate()

    # now the interlinking
    if interlinking:
        eofflags = []
        epros = []
        for e1 in endpoints:
            for e2 in endpoints:
                if e1 == e2:
                    continue
                q = multiprocessing.Queue()
                eofflags.append(q)
                logger.info('Finding inter-links between: ' + e1.url + ' and ' + e2.url)
                logger.info('==============================//=========//===============================')
                p = multiprocessing.Process(target=_get_links, args=(e1, sparqlendps[e1.url], e2, sparqlendps[e2.url], q,))
                epros.append(p)
                p.start()

        while len(eofflags) > 0:
            for q in eofflags:
                rdfmts = q.get()
                for rdfmt in rdfmts:
                    rootType = rdfmt['rootType']
                    if rootType not in dsrdfmts:
                        dsrdfmts[rootType] = rdfmt
                    else:
                        _merge_mts(rdfmt, rootType, dsrdfmts)
                eofflags.remove(q)
                break

        for p in epros:
            if p.is_alive():
                p.terminate()

    for e in sparqlendps:
        rdfmts = sparqlendps[e]
        for rdfmt in rdfmts:
            rootType = rdfmt['rootType']
            if rootType not in dsrdfmts:
                dsrdfmts[rootType] = rdfmt
            else:
                _merge_mts(rdfmt, rootType, dsrdfmts)

    templates = list(dsrdfmts.values())
    duration = time() - start
    logger.info('----- DONE in ' + str(duration) + ' seconds!-----')
    logger_wrapper.setLevel(logging.INFO)  # reset the logger
    if output is not None:
        json.dump(templates, open(output, 'w'), indent=2)
    else:
        return JSONConfig(templates)


def _collect_rdfmts_from_source(endpoint: Endpoint, tq):
    # get the typed concepts, predicates, etc. for one source
    if 'mappings' in endpoint.params:
        molecules = get_rdfmts_from_mapping(endpoint)
        tq.put(molecules)
        tq.put('EOF')
        return molecules

    concepts = _get_typed_concepts(endpoint)
    logger.info(endpoint.url + ': ' + str(concepts))
    molecules = []
    for c in concepts:
        if '^^' in c:
            continue
        logger.info(c)
        class_properties = []
        linked_to = set()
        predicates = _get_predicates(endpoint, c)
        for p in predicates:
            if 'wikiPageWikiLink' in p:
                continue
            range_ = _get_predicate_range(endpoint, c, p)
            if len(range_) > 0:
                [linked_to.add(r) for r in range_]
            class_properties.append({
                'predicate': p,
                'range': range_,
                'policies': [{'dataset': endpoint.url, 'operator': 'PR'}]
            })
        molecules.append({
            'rootType': c,
            'predicates': class_properties,
            'linkedTo': list(linked_to),
            'wrappers': [{
                'url': endpoint.url,
                'predicates': predicates,
                'urlparam': endpoint.get_params(),
                'wrapperType': 'SPARQLEndpoint'
            }]
        })

    logger.info('=================================')
    tq.put(molecules)
    tq.put('EOF')
    return molecules


def _get_typed_concepts(endpoint):
    if endpoint.types:
        return endpoint.types

    query = 'SELECT DISTINCT ?t WHERE { ?s '
    if 'wikidata' in endpoint.url:
        query += TYPING_WIKIDATA
    else:
        query += 'a'
    query += ' ?t . }'
    res_list, _ = _get_results_iter(query, endpoint)
    return [r['t'] for r in res_list if True not in [m in str(r['t']) for m in metas]]


def _get_predicates(endpoint, type_):
    query = 'SELECT DISTINCT ?p WHERE { ?s '
    if 'wikidata' in endpoint.url:
        query += TYPING_WIKIDATA
    else:
        query += 'a'
    query += ' <' + type_ + '> . ?s ?p ?pt . }'
    res_list, status = _get_results_iter(query, endpoint)
    res_list = [r['p'] for r in res_list]

    if status == -1:  # fallback - get predicates from randomly selected instances of the type
        logger.warn('giving up on ' + query)
        logger.warn('trying instances...')
        rand_inst_res = _get_predicates_of_random_instances(endpoint, type_)
        for pred in rand_inst_res:
            if pred not in res_list:
                res_list.append(pred)

    return res_list


def _get_predicates_of_random_instances(endpoint, type_):
    query = 'SELECT DISTINCT ?s WHERE { ?s '
    if 'wikidata' in endpoint.url:
        query += TYPING_WIKIDATA
    else:
        query += 'a'
    query += ' <' + type_ + '> . }'
    res_list, _ = _get_results_iter(query, endpoint, limit=100, max_tries=100, max_answers=100)
    results = []
    batches = [res_list[i:i+10] for i in range(0, len(res_list), 10)]
    for batch in batches:
        batch = ['<' + r['s'] + '>' for r in batch]
        query = 'SELECT DISTINCT ?p WHERE {\n' \
                '  VALUES ?s { ' + ' '.join(batch) + ' }\n' \
                '  ?s ?p ?pt\n}'
        res_list_batch, _ = _get_results_iter(query, endpoint)
        results.extend([r['p'] for r in res_list_batch])
    return list(set(results))


def _get_predicate_range(endpoint, type_, predicate):
    # first, get the range using rdfs:range
    query = 'SELECT DISTINCT ?range WHERE { <' + predicate + '> <http://www.w3.org/2000/01/rdf-schema#range> ?range . }'
    res_list, _ = _get_results_iter(query, endpoint)
    ranges = [r['range'] for r in res_list if True not in [m in str(r['range']) for m in metas]]

    # second, get range from instances
    query = 'SELECT DISTINCT ?range WHERE {\n' \
            '  ?s '
    if 'wikidata' in endpoint.url:
        query += TYPING_WIKIDATA
    else:
        query += 'a'
    query += ' <' + type_ + '> .\n' \
             '  ?s <' + predicate + '> ?pt .\n' \
             '  ?pt a ?range .\n}'
    res_list, _ = _get_results_iter(query, endpoint)
    ranges.extend([r['range'] for r in res_list if True not in [m in str(r['range']) for m in metas]])

    return list(set(ranges))


def _accessible_endpoints(endpoints):
    ask = 'ASK { ?s ?p ?o }'
    accessible_endpoints = []
    for e in endpoints:
        url = e.url
        val, c = contact_source(url, ask, Queue(), CONFIG)
        if c == -2:
            logger.error(url + ' --> is not accessible. Please check if this endpoint properly started!')
            sys.exit(1)
        if val:
            accessible_endpoints.append(e)
        else:
            logger.info(url + ' --> is returning empty results. Hence, will not be included in the federation!')

    return accessible_endpoints


def _get_results_iter(query: str, endpoint: Endpoint, limit: int = -1, max_tries: int = -1, max_answers: int = -1):
    offset = 0
    res_list = []
    status = 0
    num_requests = 0

    if limit == -1:
        limit = 10000

    while True:
        query_copy = query + ' LIMIT ' + str(limit) + (' OFFSET ' + str(offset) if offset > 0 else '')
        num_requests += 1
        res_queue = Queue()
        _, card = contact_source(endpoint.url, query_copy, res_queue, CONFIG)

        # if receiving the answer fails, try with a decreasing limit
        if card == -2:
            limit = limit // 2
            if limit < 1:
                status = -1
                break
            continue

        # results returned from the endpoint need to be appended to the result list
        if card > 0:
            res = res_queue.get()
            while res != 'EOF':
                res_list.append(res)
                res = res_queue.get()

        # stop if all results are retrieved or the maximum number of tries is reached
        if card < limit or (0 < max_answers <= len(res_list)) or num_requests == max_tries:
            break

        offset += limit

    return res_list, status


def _get_links(endpoint1, rdfmt1, endpoint2, rdfmt2, q):
    found = False
    for c in rdfmt1:
        for p in c['predicates']:
            if p['predicate'] == 'http://www.w3.org/1999/02/22-rdf-syntax-ns#type':
                continue
            ext_links = _get_external_links(endpoint1, c['rootType'], p['predicate'], endpoint2, rdfmt2)
            if len(ext_links) > 0:
                found = True
                c['linkedTo'].extend(ext_links)
                c['linkedTo'] = list(set(c['linkedTo']))
                p['range'].extend(ext_links)
                p['range'] = list(set(p['range']))
    if found:
        q.put(rdfmt1)
    else:
        q.put([])
    q.put('EOF')


def _get_external_links(endpoint1, root_type, predicate, endpoint2, rdfmt2):
    query = 'SELECT DISTINCT ?o WHERE { ?s '
    if 'wikidata' in endpoint1.url:
        query += TYPING_WIKIDATA
    else:
        query += 'a'
    query += ' <' + root_type + '> ; <' + predicate + '> ?o . FILTER (isIRI(?o)) }'
    e1_objects, _ = _get_results_iter(query, endpoint1, max_tries=100)
    e1_objects = [obj['o'] for obj in e1_objects]
    batches = [e1_objects[i:i + 45] for i in range(0, len(e1_objects), 45)]
    linked_to = set()
    for batch in batches:
        batch = ['<' + r + '>' for r in batch]
        query = 'SELECT DISTINCT ?c WHERE {\n' \
                '  VALUES ?o { ' + ' '.join(batch) + ' }\n' \
                '  ?s ?p ?o .\n' \
                '  ?s '
        if 'wikidata' in endpoint2.url:
            query += TYPING_WIKIDATA
        else:
            query += 'a'
        query += ' ?c .\n' \
                 '}'
        res_list, _ = _get_results_iter(query, endpoint2, max_tries=10)
        [linked_to.add(r['c']) for r in res_list]

        if len(linked_to) == len(rdfmt2):
            break
    if len(linked_to) > 0:
        logger.info(root_type + ', ' + predicate + ' --> ' + str(linked_to))
    return list(linked_to)


def _merge_mts(rdfmt, root_type, dsrdfmts):
    otherrdfmt = dsrdfmts[root_type]

    dss = {d['url']: d for d in otherrdfmt['wrappers']}

    if rdfmt['wrappers'][0]['url'] not in dss:
        otherrdfmt['wrappers'].extend(rdfmt['wrappers'])
    else:
        pps = rdfmt['wrappers'][0]['predicates']
        dss[rdfmt['wrappers'][0]['url']]['predicates'].extend(pps)
        dss[rdfmt['wrappers'][0]['url']]['predicates'] = list(set(dss[rdfmt['wrappers'][0]['url']]['predicates']))
        otherrdfmt['wrappers'] = list(dss.values())

    predicates_other = {p['predicate']: p for p in otherrdfmt['predicates']}
    predicates_this = {p['predicate']: p for p in rdfmt['predicates']}
    predicates_same = set(predicates_other.keys()).intersection(predicates_this.keys())
    for p in predicates_same:
        if len(predicates_this[p]['range']) > 0:
            predicates_other[p]['range'].extend(predicates_this[p]['range'])
            predicates_other[p]['range'] = list(set(predicates_other[p]['range']))
    preds = [predicates_other[p] for p in predicates_other]
    otherrdfmt['predicates'] = preds
    newpreds = set(list(predicates_this.keys())).difference(list(predicates_other.keys()))
    otherrdfmt['predicates'].extend([predicates_this[p] for p in newpreds])
    otherrdfmt['linkedTo'] = list(set(rdfmt['linkedTo'] + otherrdfmt['linkedTo']))


def get_rdfmts_from_mapping(endpoint):
    rdf_type = 'http://www.w3.org/1999/02/22-rdf-syntax-ns#type'
    policies = [{'dataset': endpoint.url, 'operator': 'PR'}]

    mapping_graph = Graph()
    for mapping_file in endpoint.params['mappings']:
        mapping_graph.parse(mapping_file, format='n3')

    query = 'PREFIX rr: <http://www.w3.org/ns/r2rml#> ' \
            'SELECT DISTINCT ?t ?p ?r WHERE {\n' \
            '  ?tm rr:subjectMap  ?sm. ?sm rr:class ?t .\n' \
            '  ?tm rr:predicateObjectMap ?pom .\n' \
            '  ?pom rr:predicate ?p .\n' \
            '  OPTIONAL {\n' \
            '    ?pom rr:objectMap ?om .\n' \
            '    ?om rr:parentTriplesMap ?pt .\n' \
            '    ?pt rr:subjectMap ?ptsm .\n' \
            '    ?ptsm rr:class ?r .\n' \
            '  }\n' \
            '} ORDER BY ?t ?p ?r'

    res = mapping_graph.query(query)
    metadata = {}
    for r in res:
        class_ = str(r['t'])
        if class_ not in metadata:
            metadata[class_] = {
                'predicates': {},
                'linkedTo': set()
            }
            metadata[class_]['predicates'][rdf_type] = {
                'predicate': rdf_type,
                'range': [],
                'policies': policies
            }

        predicate = str(r['p'])
        range_ = str(r['r']) if r['r'] is not None else None
        if predicate not in metadata[class_]['predicates']:
            metadata[class_]['predicates'][predicate] = {
                'predicate': predicate,
                'range': [] if range_ is None else [range_],
                'policies': policies
            }
        else:
            if range_ is not None:
                metadata[class_]['predicates'][predicate]['range'].append(range_)

        if range_ is not None:
            metadata[class_]['linkedTo'].add(range_)

    molecules = []
    for mol in metadata:
        molecules.append({
            'rootType': mol,
            'predicates': list(metadata[mol]['predicates'].values()),
            'linkedTo': list(metadata[mol]['linkedTo']),
            'wrappers': [{
                'url': endpoint.url,
                'predicates': list(metadata[mol]['predicates'].keys()),
                'urlparam': endpoint.get_params() or '',
                'wrapperType': 'SPARQLEndpoint'
            }]
        })
    return molecules



import abc


class Tree(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def instantiate(self, d):
        return

    @abc.abstractmethod
    def instantiateFilter(self, d, filter_str):
        return

    def degree(self):
        return getDegree(self.vars, self.dict)

    def __leq__(self, other):
        return (self.size < other.size or (self.size == other.size and self.degree() <= other.degree()))

    def __lt__(self, other):
        return (self.size < other.size or (self.size == other.size and self.degree() < other.degree()))

    @abc.abstractmethod
    def __eq__(self, other):
        return

    @abc.abstractmethod
    def __hash__(self):
        return

    def __ne__(self, other):
        return not self == other

    @abc.abstractmethod
    def __repr__(self):
        return

    @abc.abstractmethod
    def aux(self, n):
        return

    @abc.abstractmethod
    def aux2(self, n):
        return

    def show(self, n):
        return self.aux(n)

    def show2(self, n):
        return self.aux2(n)

    @abc.abstractmethod
    def getVars(self):
        return

    @abc.abstractmethod
    def places(self):
        return

    @abc.abstractmethod
    def constantNumber(self):
        return

    def constantPercentage(self):
        return self.constantNumber() / self.places()


class Node(Tree):

    def __init__(self, l, r, filters=None):
        if filters is None:
            filters = []
        self.left = l
        self.right = r
        self.vars = unify(l.vars, r.vars, l.dict)
        self.dict = l.dict = r.dict
        self.size = l.size + r.size
        self.filters = filters

    def instantiate(self, d):
        return Node(self.left.instantiate(d), self.right.instantiate(d))

    def instantiateFilter(self, d, filter_str):
        return Node(self.left.instantiateFilter(d, filter_str), self.right.instantiateFilter(d, filter_str))

    def __eq__(self, other):
        return ((isinstance(other, Node)) and (self.vars == other.vars) and
                (self.dict == other.dict) and (self.degree() == other.degree()) and
                (self.size == other.size) and (self.left == other.left) and
                (self.right == other.right))

    def __hash__(self):
        return hash((self.vars, self.dict, self.size, self.degree(),
                     self.left, self.right))

    def __repr__(self):
        return self.aux(" ")

    def aux(self, n):
        s = ""
        if self.left:
            s = s + n + "{\n" + self.left.aux(n + "  ") + "\n" + n + "}\n" + n + "  . \n"

        if self.right:
            s = s + n + "{\n" + self.right.aux(n + "  ") + "\n" + n + "}"
        for f in self.filters:
            s += str(f)
        return s

    def show(self, n):
        return self.aux(n)

    def show2(self, n):
        return self.aux2(n)

    def aux2(self, n):
        s = ""
        if self.left:
            s = s + n + "{\n" + self.left.aux2(n + "  ") + "\n" + n + "}\n" + n + "  UNION \n"

        if self.right:
            s = s + n + "{\n" + self.right.aux2(n + "  ") + "\n" + n + "}"
        return s

    def places(self):
        return self.left.places() + self.right.places()

    def constantNumber(self):
        return self.left.constantNumber() + self.right.constantNumber()

    def getVars(self):
        vs = []
        if self.left:
            vs = vs + self.left.getVars()
        if self.right:
            vs = vs + self.right.getVars()
        return vs


def unify(vars0, vars1, dict0):
    vars2 = set(vars0)
    for v in vars1:
        if v in vars2:
            dict0[v] = dict0[v] - 1
            if v in dict0 and dict0[v] == 0:
                del dict0[v]
                vars2.remove(v)
        else:
            vars2.add(v)
    return vars2


def getDegree(vars0, dict0):
    s = 0
    for v in vars0:
        s = s + dict0[v]
    return s


class Leaf(Tree):

    def __init__(self, s, vs, dc, filter=None):
        if filter is None:
            filter = []
        self.vars = vs
        self.dict = dc
        self.size = 1
        self.service = s
        self.filters = []
        serviceVars = s.getVars()
        for f in filter:
            vars_f = f.getVars()
            if set(serviceVars) & set(vars_f) == set(vars_f):
                self.filters.append(f)

    def __hash__(self):
        return hash((self.vars, self.dict, self.size, self.degree(),
                     self.service))

    def __repr__(self):
        return str(self.service)

    def __eq__(self, other):
        return ((isinstance(other, Leaf)) and (self.vars == other.vars) and
                (self.dict == other.dict) and (self.degree() == other.degree()) and
                (self.service == other.service))

    def instantiate(self, d):
        newvars = self.vars - set(d.keys())
        newdict = self.dict.copy()
        for c in d:
            if c in newdict:
                del newdict[c]
        return Leaf(self.service.instantiate(d), newvars, newdict)

    def instantiateFilter(self, d, filter_str):
        # print "Leaf: instantiateFilters ", self.vars, self.filters
        # print "Leaf: filterstr", filter_str
        newvars = self.vars - set(d)
        newdict = self.dict.copy()
        # print "Leaf: instantiateFilters new ", newvars, newdict
        for c in d:
            if c in newdict:
                del newdict[c]
        return Leaf(self.service.instantiateFilter(d, filter_str), newvars, newdict)

    def aux(self, n):
        return self.service.show(n)

    def aux2(self, n):
        return self.service.show2(n)

    def show(self, n):
        return self.aux(n)

    def show2(self, n):
        return self.aux2(n)

    def getInfoIO(self, query):
        subquery = self.service.getTriples()
        # print "getinfo: subquery: ", subquery
        vs = list(set(self.service.getVars()))  # - set(self.service.filters_vars)) # Modified this by mac: 31-01-2014
        # print "service", vs
        # predictVar = set(self.service.getPredVars())  # no longer needed since pred vars are included in vars
        variables = [v.lstrip("?$") for v in vs]
        if not query.args:
            projvars = vs
        else:
            # projvars = list(set([v.name for v in query.args if not v.constant]))
            projvars = []
            for arg in query.args:
                if '?ALL' in arg.getVars():
                    projvars = vs               # doubtable
                    break
                else:
                    projvars += arg.getVars()
        subvars = list((query.join_vars | set(projvars)) & set(vs))
        vars_order_by = [x for v in query.order_by for x in v.getVars() if x in vs]
        vars_group_by = [x for v in query.group_by for x in v.getVars() if x in vs]
        vars_having = list()
        if query.having:
            vars_having = [v for v in query.having.getVars() if v in vs]
        # print "subvar 1: ", subvars
        if subvars == []:
            subvars = vs
            # print "subvar 1.1: ", subvars
        filter_vars = [v for v in query.getFilterVars() if v in vs]

        subvars = list(set(subvars) | set(vars_order_by) | set(filter_vars) | set(vars_group_by) | set(vars_having))
        # print "subvar 2: ", subvars
        # This corresponds to the case when the subquery is the same as the original query.
        # In this case, we project the variables of the original query.
        if query.body.show(" ").count("SERVICE") == 1:
            subvars = list(set(projvars) | set(vars_order_by) | set(vars_group_by) | set(vars_having))
            # print "subvar 3: ", subvars
        subvars = " ".join(subvars)
        # MEV distinct pushed down to the sources
        # print 'subvars 4: ', subvars
        if query.distinct:
            d = "DISTINCT "
        else:
            d = ""

        subquery = "SELECT " + d + subvars + " WHERE {" + subquery + "\n" + query.filter_nested + "\n}"

        return (self.service.endpoint, query.getPrefixes() + subquery, set(variables))

    def getCount(self, query, vars, endpointType):
        subquery = self.service.getTriples()
        if len(vars) == 0:
            vs = self.service.getVars()
            variables = [v.lstrip("?$") for v in vs]
            vars_str = "*"
        else:
            variables = vars
            service_vars = self.service.getVars()
            vars2 = []
            for v1 in vars:
                for v2 in service_vars:
                    if (v1 == v2[1:]):
                        vars2.append(v2)
                        break
            if len(vars2) > 0:
                vars_str = " ".join(vars2)
            else:
                vars_str = "*"

        d = "DISTINCT "
        if (endpointType == "V"):
            subquery = "SELECT COUNT " + d + vars_str + "  WHERE {" + subquery + "\n" + query.filter_nested + "}"
        else:
            subquery = "SELECT ( COUNT (" + d + vars_str + ") AS ?cnt)  WHERE {" + subquery + "\n" + query.filter_nested + "}"

        return self.service.endpoint, query.getPrefixes() + subquery

    def getVars(self):
        return self.service.getVars()

    def places(self):
        return self.service.places()

    def constantNumber(self):
        return self.service.constantNumber()


def sort(lss):
    lo = []
    while not (lss == []):
        m = 0
        for i in range(len(lss)):
            if lss[i].constantPercentage() > lss[m].constantPercentage():
                m = i
        lo.append(lss[m])
        lss.pop(m)
    return lo


def createLeafs(lss, filters=None):
    if filters is None:
        filters = []
    d = dict()
    for s in lss:
        l = s.getVars()
        l = set(l)
        for e in l:
            d[e] = d.get(e, 0) + 1
    el = []
    for e in d:
        d[e] = d[e] - 1
        if d[e] <= 0:
            el.append(e)
    for e in el:
        del d[e]
    ls = []
    lo = sort(lss)

    for s in lo:
        e = set()
        l = s.getVars()
        for v in l:
            if v in d:
                e.add(v)
        ls.append(Leaf(s, e, d, filters))

    return (d, ls)


# def shareAtLeastOneVar(l, r):
#     if isinstance(l, Leaf) and isinstance(r, Leaf):
#         ls = l.service.triples[0].subject.name
#         lo = [t.theobject.name for t in l.service.triples]
#         rs = r.service.triples[0].subject.name
#         ro = [t.theobject.name for t in r.service.triples]
#         return len(l.vars & r.vars) > 0 or ls in ro or rs in lo
#
#     return len(l.vars & r.vars) > 0


def shareAtLeastOneVar(l, r):
    return len(l.vars & r.vars) > 0


def sortedInclude(l, e):
    for i in range(0, len(l)):
        if e < l[i]:
            l.insert(i, e)
            return
    l.insert(len(l), e)


def updateFilters(node, filters):
    if isinstance(node, Leaf):
        return Leaf(node.service, node.vars, node.dict, node.filters + filters)
    elif isinstance(node, Node):
        return Node(node.left, node.right, node.filters + filters)


def makeNode(l, r, filters=None):
    if filters is None:
        filters = []
    if l.constantPercentage() > r.constantPercentage():
        n = Node(l, r, filters)
    else:
        n = Node(r, l, filters)
    return n


def makeBushyTree(ss, filters=None):
    if filters is None:
        filters = []
    (d, pq) = createLeafs(ss, filters)

    others = []
    while len(pq) > 1:
        done = False
        l = pq.pop(0)  # heapq.heappop(pq)
        lpq = pq  # heapq.nsmallest(len(pq), pq)

        for i in range(0, len(pq)):
            r = lpq[i]

            if shareAtLeastOneVar(l, r):
                pq.remove(r)
                n = makeNode(l, r, filters)
                pq.append(n)  # heapq.heappush(pq, n)
                done = True
                break
        if not done:
            others.append(l)

    if len(pq) == 1:
        for e in others:
            pq[0] = makeNode(pq[0], e, filters)
        return pq[0]
    elif others:
        while len(others) > 1:
            l = others.pop(0)
            r = others.pop(0)

            n = Node(l, r, filters)
            others.append(n)
        if others:
            return others[0]
        return None


def makeNaiveTree(ss, filters=None):
    if filters is None:
        filters = []
    (_, pq) = createLeafs(ss)
    while len(pq) > 1:
        l = pq.pop(0)
        r = pq.pop(0)

        n = makeNode(l, r, filters)
        pq.append(n)

    if len(pq) == 1:
        return pq[0]
    else:
        return None


def makeLLTree(ss, filters=None):
    if filters is None:
        filters = []
    (_, pq) = createLeafs(ss)
    while len(pq) > 1:
        l = pq.pop(0)
        r = pq.pop(0)

        n = makeNode(l, r, filters)
        pq.insert(0, n)

    if len(pq) == 1:
        return pq[0]
    else:
        return None



"""
Created on Jul 10, 2011

Implements a Hash Optional operator.
The intermediate results are represented as a Queue, but it doesn't
work until all the tuples are arrived.

@author: Maribel Acosta Deibe
"""

from time import time
from DeTrusty.Operators.Optional import Optional
from .OperatorStructures import Table, Record
from multiprocessing import Queue


class HashOptional(Optional):

    def __init__(self, vars_left, vars_right):
        self.left_table  = Table()
        self.right_table = Table()
        self.results     = []
        self.bag         = []
        self.results     = []
        self.vars_left   = set(vars_left)
        self.vars_right  = set(vars_right)
        self.vars        = list(self.vars_left & self.vars_right)

    def instantiate(self, d):
        newvars_left = self.vars_left - set(d.keys())
        newvars_right = self.vars_right - set(d.keys())
        return HashOptional(newvars_left, newvars_right)

    def execute(self, qleft, qright, out, processqueue=Queue()):
        # Executes the Hash Optional.
        self.left = []
        self.right = []
        self.qresults = out

        # Initialize tuples.
        tuple1 = None
        tuple2 = None

        # Get the tuples from the queues.
        while (not(tuple1 == "EOF") or not(tuple2 == "EOF")):
            # Try to get tuple from left queue.
            if not(tuple1 == "EOF"):
                try:
                    tuple1 = qleft.get(False)
                    self.left.append(tuple1)
                except Exception:
                    # This catch:
                    # Empty: in tuple2 = self.left.get(False), when the queue is empty.
                    pass

            # Try to get tuple from right queue.
            if not(tuple2 == "EOF"):
                try:
                    tuple2 = qright.get(False)
                    self.right.append(tuple2)
                except Exception:
                    # This catch:
                    # Empty: in tuple2 = self.right.get(False), when the queue is empty.
                    pass

        # Get the variables to join.
        if ((len(self.left) > 1) and (len(self.right) > 1)):
            # Iterate over the lists to get the tuples.
            while ((len(self.left) > 1) or (len(self.right) > 1)):
                if len(self.left) > 1:
                    tuple1 = self.left.pop(0)
                    self.bag.append(tuple1)
                    self.insertAndProbe(tuple1, self.left_table, self.right_table)
                if len(self.right) > 1:
                    tuple2 = self.right.pop(0)
                    self.insertAndProbe(tuple2, self.right_table, self.left_table)

        # This is the optional.
        res_right = {}
        for var in self.vars_right:
            res_right.update({var: ''})
        for tuple in self.bag:
            res = res_right
            res.update(tuple)
            self.results.append(res)

        # Put all the results in the output queue.
        while self.results:
            self.qresults.put(self.results.pop(0))

        # Put EOF in queue and exit.
        self.qresults.put("EOF")

    def insertAndProbe(self, tuple, table1, table2):
        # Insert the tuple in its corresponding partition and probe.

        # Get the attribute(s) to apply hash.
        att = ''
        for var in self.vars:
            att = att + tuple[var]
        i = hash(att) % table1.size;

        # Insert record in partition.
        record = Record(tuple, time(), 0)
        table1.insertRecord(i, record)

        # Probe the record against its partition in the other table.
        self.probe(record, table2.partitions[i], self.vars)

    def probe(self, record, partition, var):
        # Probe a tuple if the partition is not empty.
        if partition:
            # For every record in the partition, check if it is duplicated.
            # Then, check if the tuple matches for every join variable.
            # If there is a join, concatenate the tuples and produce result.
            # If there is no join, concatenate the tuple with an empty tuple.
            for r in partition.records:
                if self.isDuplicated(record, r):
                    break

                for v in var:
                    join = True
                    if record.tuple[v] != r.tuple[v]:
                        join = False
                        break

                # This is the join.
                if join:
                    res = record.tuple.copy()
                    res.update(r.tuple)
                    self.results.append(res)

                    try:
                        self.bag.remove(r.tuple)
                    except ValueError:
                        pass

                    try:
                        self.bag.remove(record.tuple)
                    except ValueError:
                        pass

    def isDuplicated(self, record1, record2):
        # Verify if a two tuples has been already probed.
        return not record1.ats >= record2.ats
